# Copyright (c) 2005, Henry Ludemann
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
#  * Neither the name of the organization nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""cmdline - A powerful easy to use option parser

Provides a high level command line interface to an existing python method, it
is designed to be quick and easy to use. It fulfills similar functionality to
the optparse module. Functionality is provided through the Command class in
the cmdline.Command module.
"""

from cmdline.basecommand import BaseCommand
import inspect
import sys

class Command(BaseCommand):
    """ A class to invoke a function depending on command line arguments """
    def __init__(self, method, description, post_description="",
            output=sys.stdout.write, error=sys.stderr.write):
        """Construct a command object

        method -- The method that will be called given successful parsing
        description -- A textual description of what the command does (used
            when printing the help)
        post_description -- Text that follows the help (for example, 'Report
           bugs to blah@example.com')
        """
        if inspect.isclass(method):
            raise NotImplementedError("cmdline cannot be used to initialise a class ('%s')" % method)

        BaseCommand.__init__(self, description, post_description, output, error)
        self._method = method

    def _invoke(self, values):
        params = self._get_params(values)
        return self._method(*params)

    def _validate_param_name(self, name):
        assert name is not None
        args, varargs, varkw, defaults = inspect.getargspec(self._method)
        if not name in args:
            raise ValueError("%s isn't a valid parameter to %s! Options are: %s" % \
                    (name, self._method.__name__, ' '.join(args)))

    def _is_required(self, name):
        """ Check to see if we have a default value for a given parameter """
        args, varargs, varkw, defaults = inspect.getargspec(self._method)
        if defaults is None:
            defaults = []
        return args.index(name) < len(args) - len(defaults)

    def _get_params(self, values):
        """ Get a parameter list for this method """

        # Create a list of method parameter values, using the named value
        # if it exists, otherwise using the default parameter
        method_args, varargs, varkw, defaults = inspect.getargspec(self._method)
        if defaults is None:
            # When there are no default parameters, we want to handle it
            # correctly
            defaults = []

        parameters = []
        for index, arg in enumerate(method_args): 
            if index == 0 and inspect.ismethod(self._method) and self._method.im_self:
                # The first value is the instance of the object, and as such
                # we don't need to override it (calling the method object fills
                # in this value)
                pass
            elif values.has_key(arg):
                parameters.append(values[arg])
            else:
                # The base command should have validated that we have all
                # required arguments by this stage...
                default_index = index - (len(method_args) - len(defaults))
                parameters.append(defaults[default_index])
        return parameters


class ConfigSetter(BaseCommand):
    """ Class to set attributes in an object depending on command line arguments.
    
        When invoke is called, returns None if command line was successfully
        parsed. """
    def __init__(self, config, description, post_description="",
            output=sys.stdout.write, error=sys.stderr.write):
        BaseCommand.__init__(self, description, post_description, output, error)
        assert config is not None
        self._config = config

    def _validate_param_name(self, full_name):
        assert full_name
        obj = self._config
        for name in full_name.split('.'):
            if not hasattr(obj, name):
                raise ValueError("'%s' isn't an attribute of '%s'!" % (name, obj))
            obj = getattr(obj, name)

    def _invoke(self, values):
        for full_name, value in values.iteritems():
            names = full_name.split('.')
            obj = self._config
            for name in names[:-1]:
                obj = getattr(obj, name)
            setattr(obj, names[-1], value)
