# Copyright (c) 2005, 2007, Henry Ludemann
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
#  * Neither the name of the organization nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import getopt
import textwrap

from cmdline.flag import Flag
from cmdline.argument import Argument, ArgumentError
from cmdline.commandargument import CommandArgument


class _ParamError(Exception):
    """ Exception raised when we can't parse the command line arguments """
    pass

class BaseCommand(object):
    """ Base class used for command line parsing.
    
    This class will not be instantiated directly; only derived classes are
    useful.

    Derived classes have to implement BaseCommand._invoke, and can implement
    BaseCommand._validate_param_name and BaseCommand._is_required.
    """

    def __init__(self, description, post_description="", \
            output=sys.stdout.write, error=sys.stderr.write):
        self._description = description
        self._post_description = post_description
        self._short_flags = ''
        self._long_flags = []
        self._flags = []
        self._arguments = []

        self._output = output
        self._error = error

    def add_flag(self, option, long_option, description, param_name, 
                 get_argument=None, argument_name=None):
        """Add a flag (that can have a value)

        Either option or long_option can be None (but not both)
        """
        assert param_name is not None
        assert description
        if option is not None and option.startswith('-'):
            raise ValueError("Options must not start with the '-' " \
                    "character ('%s')" % option)
        if long_option is not None and long_option.startswith('-'):
            raise ValueError("Options must not start with the '-' " \
                    "character ('%s')" % long_option)

        has_argument = get_argument is not None
        self._validate_param_name(param_name)
        self._add_options(option, long_option, has_argument)

        # Add the flag to the array 
        argument = None
        if has_argument:
            if not argument_name:
                argument_name="VALUE"
            argument = Argument(argument_name, get_argument)
        flag = Flag(param_name, option, long_option, description, argument)
        self._flags.append(flag)

    def add_argument(self, name, description, param_name, get_argument=str):
        """Add an argument to be read from the command line

        Takes the name of the argument, the name of the parameter it will be
        used for, and a function to convert the argument to the correct type
        """
        assert param_name is not None
        assert description
        self._validate_param_name(param_name)

        argument = CommandArgument(name, description, get_argument, param_name) 
        self._arguments.append(argument)

    def _add_options(self, option, long_option, has_value):
        """Add the flags to the option lists, and ensure they are unique"""
        if option:
            if option in self._short_flags:
                raise ValueError(option + " option has allready been used!")

            short_flag = option
            if has_value:
                short_flag = short_flag + ':'
            self._short_flags = self._short_flags + short_flag

        if long_option:
            for current_long_option in self._long_flags:
                if long_option == current_long_option.strip('='):
                    raise ValueError("%s option has allready been added!" % long_option)
            long_flag = long_option
            if has_value:
                long_flag = long_flag + "="
            self._long_flags.append(long_flag)

    def _validate_param_name(self, param_name):
        """Can be overridden by derived classes to validate parameter names. """
        return

    def _is_required(self, param_name):
        """ Can be overridden by derived classes to check for required parameters. """
        return False

    def _get_flag(self, flag):
        "Get a given named flag"
        for flag_obj in self._flags:
            if flag_obj.is_flag(flag):
                return flag_obj
        raise ValueError("Couldn't find flag %s" % flag)

    def _get_lookup_values(self, opts, args):
        """Get a list of parameters to be used when invoking the method

        Uses the default parameter values, flag list, and arguments.
        
        @return A list of (param name, value) tuples.
        @raise _ParamError is raised if we can't detect parameters. 
        """
        # Create a dictionary of parameters by name to value
        values = {}
        for flag, value in opts.iteritems():
            flag_obj = self._get_flag(flag)
            try:
                value = flag_obj.get_value(value)
            except ArgumentError, ex:
                raise _ParamError(ex)
            values[flag_obj.param_name] = value

        # Add all arguments to the dictionary
        if len(args) > len(self._arguments):
            raise _ParamError("Too many arguments! Expected %i got %i" % \
                (len(self._arguments), len(args)))

        for index, arg_string in enumerate(args):
            argument = self._arguments[index]
            try:
                value = argument.get_value(arg_string)
            except ArgumentError, ex:
                raise _ParamError(ex)
            values[argument.param_name] = value

        # Check that all required values have been set
        for flag in self._flags:
            if flag.param_name not in values:
                if self._is_required(flag.param_name):
                    raise _ParamError("Required flag '%s' must be set!" % flag)
        for argument in self._arguments:
            if argument.param_name not in values:
                if self._is_required(argument.param_name):
                    raise _ParamError("Required argument '%s' must be set!" % argument.name)

        return values

    def _help_flag(self):
        return Flag(None, 'h', 'help', 'Show this help page')

    def _display_usage(self, command):
        """Print the usage of the command"""
        self._output("Usage: %s [OPTION]... %s\n" % (command, \
                " ".join(a.name for a in self._arguments)))
        wrapper = textwrap.TextWrapper()
        self._output("%s\n" % wrapper.fill(self._description))

        # Print the arguments
        wrapper.subsequent_indent = ' ' * 24
        if len(self._arguments) > 0:
            self._output("Arguments\n")
            for argument in self._arguments:
                wrapper.initial_indent = "  %-21s " % (argument.name)
                self._output("%s\n" % wrapper.fill(argument.description))
            self._output("\n")

        # Sort the flags alphabetically, then print them out
        wrapper.subsequent_indent = ''
        wrapper.initial_indent = ''
        self._output("%s\n" % wrapper.fill("Mandatory arguments to long flags are " \
            "mandatory for short options too"))
        flags = self._flags[:] + [self._help_flag()]
        flags.sort()
        wrapper.subsequent_indent = ' ' * 24
        for flag in flags:
             wrapper.initial_indent = "  %-21s " % flag
             self._output("%s\n" % wrapper.fill(flag.description))

        if len(self._post_description) > 0:
            wrapper.initial_indent = ''
            wrapper.subsequent_indent = ''
            self._output("\n%s\n" % wrapper.fill(self._post_description))

    def invoke(self, command=sys.argv[0], args=sys.argv[1:]): 
        """Run the command with the given arguments.

        If an error with the parsing occurs, it prints an error before
        returning with a value of 1. If the target function is called (if
        the parameters parsed correctly) that return value will be returned.
        The command variable is used to print a useful 'usage' line.
        """
        short_flags = 'h' + self._short_flags
        long_flags = self._long_flags[:] + ['help']
        try:
            opts, args = getopt.getopt(args, short_flags, 
                                       long_flags)
        except getopt.GetoptError, ex:
            # print help information and exit:
            self._error("Error running %s %s\n" % (command, ' '.join(args)))
            self._error("%s\n" % ex)
            self._error("Run '%s --help' for more information\n" % command)
            return 1

        # Change the options to a dictionary, as this is more useful. Note
        # that this inhibits the reuse of flags.
        opts = dict(opts)

        # Do a special check for the internal help flags
        for opt in opts:
            if self._help_flag().is_flag(opt):
                self._display_usage(command)
                return 0

        # Get the parameters from the command line arguments
        try:
            values = self._get_lookup_values(opts, args)
        except _ParamError, ex:
            self._error("%s\n" % ex)
            self._error("Run '%s --help' for more information\n" % command)
            return 1

        return self._invoke(values)

    def _invoke(self, values):
        """ Class to be implemented by the derived class.
        
        @return The return value will be reused from BaseCommand.invoke.
        """
        raise NotImplementedError()

