# Copyright (c) 2005, Henry Ludemann
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
#  * Neither the name of the organization nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

class Flag(object):
    "Class to hold information about a command line flag"
    
    def __init__(self, param_name, short_flag=None, long_flag=None, 
            desc="", argument = None):
        """Construct a flag object

        Arguments
        param_name -- Parameter to be set to the result of this flag
        short_flag -- A single character representing the 'short' command
            line option. Can be None if no short flag will be used.
        long_flag -- A string representing the 'long' command line option.
            Can be None if no long flag will be used.
        desc -- Description of what this flag does
        argument -- An optional Argument object, if the flag takes an
            argument
        """
        if short_flag is None and long_flag is None:
            raise Exception, "Either the short flag or the long flag must " \
                "be set for param %s"%(param_name)
        self._short_flag = short_flag
        self._long_flag = long_flag
        self._description = desc
        self._param_name = param_name
        self._argument = argument

    description = property(lambda self:self._description, 
        doc="Description of what this flag does")

    def is_flag(self, flag_text):
        """Does this flag match a given flag string"""
        if self._short_flag is not None and \
            "-" + self._short_flag == flag_text:
            return 1
        if self._long_flag is not None and \
            "--" + self._long_flag == flag_text:
            return 1
        return 0
        
    def get_value(self, arg_value):
        """Get the value of the parameter, given that the flag has been set.

        arg_value is the value given to the flag, or null if it doesn't have
        one.
        """
        if self._argument is None:
            # This argument doesn't take a parameter, so take the flags 
            # presence to mean the flags value is true.
            if (arg_value is not None) and (arg_value != ""):
                raise ValueError, "Flag %s recieved argument %s when it " \
                    "didn't expect one!"%(str(self), str(arg_value))
            return True
        else:
            return self._argument.get_value(arg_value)

    param_name = property(lambda self:self._param_name, doc="The name of " \
        "the parameter this flag sets")

    def get_flag_name(self):
        """Get the name of a flag

        If the short name exists, it returns that. Otherwise returns the long 
        flag.
        """
        if self._long_flag is not None:
            return self._long_flag
        return self._short_flag

    def __cmp__(self, other):
        return cmp(self._short_flag or self._long_flag[0], other._short_flag or other._long_flag[0])

    def __str__(self):
        if self._short_flag and self._long_flag:
            result = '-%s, --%s'%(self._short_flag, self._long_flag)
        elif self._short_flag:
            result = '-%s'      %(self._short_flag)
        else:
            result = '    --%s'%(self._long_flag)

        # If we have an argument, add it to the string
        if self._argument:
            if self._long_flag:
                result = "%s=%s"%(result, self._argument.name)
            else:
                result = "%s %s"%(result, self._argument.name)
        return result
