# Copyright (c) 2005, Henry Ludemann
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 
#  * Neither the name of the organization nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import unittest
from cmdline import Command

class TestCommand(unittest.TestCase):

   def setUp(self):
       self.was_called = 0
       self.was_flag_set = 0
       self.flag_with_value = 0

   def _some_method(self, someFlag=5, isFlagSet=0, flag_with_value=0, blah=3):
       self.was_called = True
       self.was_flag_set = isFlagSet
       self.flag_with_value = flag_with_value
       return someFlag

   def _requires_value(self, value):
       self.was_called = True
       return value

   def _no_defaults(self):
       "A method to be invoked that doesn't have defaults"
       self.was_called = True
       pass

   def test_no_flags(self):
       "Test that we invoke command methods correctly without parameters"
       command = Command(self._some_method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
       result = command.invoke(args=[])
       self.assert_(self.was_called)
       self.assertEqual(5, result)
       self.assert_(not self.was_flag_set)
       self.assertEqual(0, self.flag_with_value)

   def test_help_command(self):
       """Test that can invoke using the help command""" 
       command = Command(self._some_method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
       command.add_flag('s', None, 'Set some flag', 'someFlag')
       command.add_flag(None, 'isflagset', 'Set some other flag', 'isFlagSet')
       command.add_flag('f', 'flagvalue', 'Set the value of some flag', 'flag_with_value', int, 'VALUE')
       command.add_argument('BLAH', 'Some argument', 'blah')
       result = command.invoke(args=['-h'])
       self.assert_(not self.was_called)
       self.assertEqual(0, result)

   def test_unknown_command(self):
       "Test that we fail when we come across an unknown command"
       command = Command(self._some_method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None) 
       result = command.invoke(args=['--badoption'])
       self.assert_(not self.was_called)
       self.assertEqual(1, result)

   def test_unused_flag(self):
       "Test that flags are optional; they don't need to be used" 
       command = Command(self._some_method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
       command.add_flag('s', 'setflag', 'Set the test flag', 'isFlagSet')
       result = command.invoke(args=[])
       self.assert_(self.was_called)
       self.assert_(not self.was_flag_set)

   def test_used_flag(self):
       "Test that flags are correctly passed through" 
       command = Command(self._some_method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
       command.add_flag('s', 'setflag', 'Set the test flag', 'isFlagSet')
       result = command.invoke(args=['-s'])
       self.assert_(self.was_flag_set)
       
   def test_flag_with_unexpected_value(self):
       "Test that flags with unexpected values fail"
       command = Command(self._some_method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
       command.add_flag('s', 'setflag', 'Set the test flag', 'isFlagSet')
       result = command.invoke(args=['--setflag=5'])
       self.assertEqual(1, result)
       self.assert_(not self.was_called)

   def test_flag_value(self):
       "Test that we can have flags with a value"
       command = Command(self._some_method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
       command.add_flag('v', 'value', 'Set the test flag', 'flag_with_value', int, "SOMEVALUE")
       command.invoke(args=['-v 63'])
       self.assertEqual(63, self.flag_with_value)

   def test_readded_options(self):
       "Test that an exception is raised when an option is set twice"
       command = Command(self._some_method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
       command.add_flag('v', 'value', 'Set the test flag', 'flag_with_value', int, "SOMEVALUE")
       # Test a readded short option
       self.assertRaises(ValueError, command.add_flag, 'v', 'val', 'Set the value flag', 'flag_with_value', int, "SOMEVALUE")
       # Test a readded long option
       self.assertRaises(ValueError, command.add_flag, 'x', 'value', 'Set the value flag', 'flag_with_value', int, "SOMEVALUE")

   def test_flag_with_bad_param(self):
       "Test that we get an exception when a unknown parameter name is used"
       command = Command(self._some_method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
       self.assertRaises(ValueError, command.add_flag, 'v', 'val', 'Set the value flag', 'missingParameter', int)

   def test_argument(self):
       "Test that we can pass arguments"
       command = Command(self._some_method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
       command.add_argument('RETURN_VALUE', 'Value to be returned', 'someFlag', int)
       self.assertEqual(5, command.invoke(args=['5']))

   def test_bad_argument(self):
       "Test that we fail to invoke when a bad argument is passed in"
       command = Command(self._some_method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
       command.add_argument('RETURN_VALUE', 'Value to be returned', 'someFlag', int)
       self.assertEqual(1, command.invoke(args=['chicken']))

   def test_bad_argument_param_name(self):
       "Test that we get an exception when we add an argument with a bad parameter"
       command = Command(self._some_method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
       self.assertRaises(ValueError, command.add_argument, 'SOME_ARG', 'Set the value flag', 'missingParameter', int)

   def test_lambda_method(self):
       "Test that we can call lambda methods from the command object"
       method = lambda lambdaParam:self._some_method(lambdaParam)
       command = Command(method, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
       command.add_flag('r',None, 'Set the return value', 'lambdaParam', str, 'RETURN_VALUE')
       self.assertEqual('bobs uncle', command.invoke(args=['-r','bobs uncle']))
        
   def test_no_defaults(self):
        "Test that we can invoke on methods that don't have defaults"
        command = Command(self._no_defaults, "Hey, I'm a description!", output=lambda a:None, error=lambda a:None)
        command.invoke(args=[])
        self.assert_(self.was_called)

   def test_options_dont_start_with_dask(self):
        command = Command(self._no_defaults, "", output=lambda a:None, error=lambda a:None)
        self.assertRaises(ValueError, command.add_flag, '-a', None, 'desc', 'someFlag')
        self.assertRaises(ValueError, command.add_flag, None, '--apple', 'desc', 'someFlag')

   def test_missing_required_argument(self):
       error = []
       command = Command(self._requires_value, "", output=lambda a:None, error=lambda msg:error.append(msg))
       command.add_argument('BLAH', 'Some argument', 'value', int)
       result = command.invoke(args=[])
       self.assert_(not self.was_called)
       self.assertTrue("Required argument 'BLAH' must be set!" in ''.join(error))

   def test_missing_required_flag(self):
       error = []
       command = Command(self._requires_value, "", output=lambda a:None, error=lambda msg:error.append(msg))
       command.add_flag('r', None, 'Some flag', 'value')
       result = command.invoke(args=[])
       self.assert_(not self.was_called)
       self.assertTrue("Required flag '-r' must be set!" in ''.join(error))

   def test_help_sort_order(self):
       output = []
       flags = ['r', 'carrot', 'b']
       command = Command(self._some_method, "", output=lambda msg:output.append(msg))
       command.add_flag(flags[0], None, 'Some flag', 'someFlag')
       command.add_flag(None, flags[1], 'Some flag', 'isFlagSet')
       command.add_flag(flags[2], None, 'Some flag', 'flag_with_value')
       command.invoke(args=['--help']) 

       order = []
       for i, line in enumerate(output):
           for flag in flags:
               if "-%s" % flag in line:
                   order.append(flag)
                   break

       self.assertEqual(3, len(order))
       self.assertEqual("b", order[0])
       self.assertEqual("carrot", order[1])
       self.assertEqual("r", order[2])
