import pycro.common.config
from pycro.backend import default_config

# The base address is the address where our code starts. So when
# we want to perform an call to an 'absolute' address, it has to
# be relative to this position.
BASE_ADDRESS = 0x0100

# Define some constants
class Register:
    AX = 0
    CX = 1
    DX = 2
    BX = 3
    SP = 4
    BP = 5
    SI = 6
    DI = 7

class Register8Bit:
    AL = 0
    CL = 1
    DL = 2
    BL = 3
    AH = 4
    CH = 5
    DH = 6
    BH = 7

class SegmentRegister:
    ES = 0
    CS = 1
    SS = 2
    DS = 3
    FS = 4
    GS = 5

class Mod:
    DISPLACE_ZERO_BYTES = 0
    DISPLACE_ONE_BYTE = 1
    DISPLACE_TWO_BYTES = 2
    DIRECT_REGISTER = 3

class R_M:
    """ Values for the r_m field when the mod field isn't DIRECT_REGISTER """
    BX_SI = 0
    BX_DI = 1
    BP_SI = 2
    BP_DI = 3
    SI = 4
    DI = 5
    BP = 6
    BX = 7

class ConditionCode:
    NO_OVERFLOW = 1
    CARRY = 2
    EQUAL = 4
    NOT_EQUAL = 5
    GREATER = 15
    GREATER_EQUAL = 13
    LESSER = 12
    LESSER_EQUAL = 14

class CodeGen:
    """ Class responsible for inserting machine code into the 
        output """
    def __init__(self, config=default_config()):
        self._output = []
        self._stack_size = 0
        self._post_steps = {}
        self._new_post_steps = []
        self._final_build_steps = []
        self._global_variables = []
        self.config = config

    def create_output(self):
        while len(self._new_post_steps) > 0:
            post_step = self._new_post_steps.pop()
            post_step()

        for final_step in self._final_build_steps:
            final_step()

        # We put the global variables immediately after the code in memory
        initial_size = len(self._output)
        variable_address = BASE_ADDRESS + len(self._output)
        for var in self._global_variables:
            var.resolve(variable_address)
            variable_address += 2
        assert initial_size == len(self._output)
        return self._output

    def get_post_step(self, post_step):
        """
        Get an instance of a post build step.

        The post_step will by construct with the instance of the code
        generator. When the post step is run, the instance returned
        will itself be called.

        If the post step is queried after it has allready been run, it
        will run again.

        Post steps may add more instructions (and add more post steps).
        """
        if post_step not in self._post_steps:
            instance = post_step(self)
            self._post_steps[post_step] = instance
        instance = self._post_steps[post_step]
        if instance not in self._new_post_steps:
            self._new_post_steps.append(instance)
        return instance

    def add_final_step(self, final_step):
        """ Register an object to be called just before the final output is
            generated. Callback may not add more instructions, but they
            may modify existing ones. """
        self._final_build_steps.append(final_step)

    def add_global_variable(self, var):
        """ Register an object to be called before the final output is created.
            
            Callback may modify existing instructions, but
            no more instructions can be added. """
        self._global_variables.append(var)

    def add_instr(self, code):
        """ Add a new instruction """
        assert 0 <= code < 256
        self._output.append(code)

    def set_instr(self, offset, code):
        """ Set the value of an existing instruction """
        assert 0 <= offset < self.get_code_offset()
        assert 0 <= code < 256
        self._output[offset] = code

    def get_code_offset(self):
        return len(self._output)

    def get_free_offset(self):
        """ Get the address of the first unused part of memory """
        return self.get_code_offset() + len(self._global_variables) * 2

    def get_stack_size(self):
        """ Get the offset of the current item in the stack """
        return self._stack_size

    def set_stack_size(self, stack_size):
        """ Set the size of the stack (the number of bytes that the current
            function can reference). """
        assert stack_size >= 0
        self._stack_size = stack_size

    def push(self, register):
        """ Push the current result onto the stack """
        PUSH = 0x50
        self.add_instr(PUSH + register)
        self._stack_size += 2

    def pop(self, register):
        """ Pop the value currently in the stack into the register """
        POP = 0x58
        self.add_instr(POP + register)
        self._stack_size -= 2
        assert self._stack_size >= 0

    def add_rm_instr(self, code, mod, spare, r_m, displacement = None):
        assert 0 <= spare < 8

        self.add_instr(code)
        self.add_instr(mod << 6 | spare << 3 | r_m)
        if mod == Mod.DIRECT_REGISTER:
            pass
        elif mod == Mod.DISPLACE_ONE_BYTE:
            assert displacement is not None
            self.add_instr(displacement)
        elif mod == Mod.DISPLACE_ZERO_BYTES:
            assert displacement is None
        else:
            assert 0

    def add_binary_instr(self, code):
        """ Add an 'rm' instruction that operates using two operators. """
        self.pop(Register.BX)
        self.pop(Register.AX)
        self.add_rm_instr(code, Mod.DIRECT_REGISTER, Register.BX, Register.AX)
        self.push(Register.AX)

    def push_data(self, data):
        """ Push fixed data onto the stack """
        # We allow values up to 2^16 to support unsigned operations (eg:
        # when using 0xffff)
        assert -32768 <= data < 65536
        self.add_instr(0x68) # Push 
        self.add_instr(data & 0x00ff) # LSB
        self.add_instr((data & 0xff00) >> 8) # MSB
        self._stack_size += 2

    def pop_data(self, length=2):
        """ Pop the data from the stack (and ignore the value).
        
            Does not modify any registers other then SP. """
        assert self._stack_size >= length

        ADD_BYTE = 0x83
        self.add_rm_instr(ADD_BYTE, Mod.DIRECT_REGISTER, 0, Register.SP)
        self.add_instr(length)
        self._stack_size -= length

    def swap_top_of_stack(self):
        """ Swap the top two items on top of the stack """
        self.pop(Register.AX)
        self.pop(Register.BX)
        self.push(Register.AX)
        self.push(Register.BX)
