import pycro.backend.codegen as cgen
import pycro.backend.operations.integer as integer

class DifferingStackSizesError(Exception):
    """
    Error raised when the same part of code is reached with differing stack sizes.
    """
    def __init__(self, stack_size, jump_stack_size):
        self._stack_size = stack_size
        self._jump_stack_size = jump_stack_size

    def __str__(self):
        return "Jump location stack size was %i, expected %i" % (self._stack_size, self._jump_stack_size)

class UnresolvedLabelError(Exception):
    """
    Error raised when a label is referenced but never resolved.
    """
    pass

class Label:
    """
    A label is a statically referencable position in the code.

    A label has a single position in the code, but can be referenced
    multiple times.
    """
    def __init__(self, codegen):
        self._codegen = codegen
        self._offset = None
        self._stack_size = None
        self._references = []
        self._codegen.add_final_step(self._ensure_resolved)

    def resolve(self):
        """
        Resolve references to this label to the current code position.

        Resolving the same label multiple times is an error.
        """
        assert self._offset is None
        self._offset = self._codegen.get_code_offset()
        self._stack_size = self._codegen.get_stack_size()
        self._resolve()

    def _resolve(self):
        if self._offset is None:
            return

        for offset, stack_size in self._references:
            if stack_size != self._stack_size:
                raise DifferingStackSizesError(self._stack_size, stack_size)

            ip_addr = offset + 2
            jump_size = self._offset - ip_addr
            self._codegen.set_instr(offset, jump_size & 0xff) # LSB
            self._codegen.set_instr(offset + 1, (jump_size >> 8) & 0xff) # MSB
        self._references = []

    def _ensure_resolved(self):
        if len(self._references) > 0:
            raise UnresolvedLabelError()

    def jump(self):
        """
        Jump to the label.
        """
        self._references.append((self._codegen.get_code_offset() + 1, self._codegen.get_stack_size()))

        JMP = 0xE9
        self._codegen.add_instr(JMP)
        self._codegen.add_instr(0)
        self._codegen.add_instr(0)
        self._resolve()

    def jump_if_false(self):
        """
        Pop an item from the stack, and jump to the label if it
        is False.
        """
        self._conditional_jump(cgen.ConditionCode.EQUAL)

    def jump_if_true(self):
        """
        Pop the top item from the stack, and jump to the label if
        it is True.
        """
        self._conditional_jump(cgen.ConditionCode.NOT_EQUAL)

    def _conditional_jump(self, condition_code):
        self._codegen.pop(cgen.Register.AX)

        TEST = 0x85
        self._codegen.add_rm_instr(TEST, cgen.Mod.DIRECT_REGISTER, cgen.Register.AX, cgen.Register.AX)

        self._references.append((self._codegen.get_code_offset() + 2, self._codegen.get_stack_size()))

        JMP_CC = 0x80
        self._codegen.add_instr(0x0F)
        self._codegen.add_instr(JMP_CC + condition_code)
        self._codegen.add_instr(0)
        self._codegen.add_instr(0)
        self._resolve()
