import logging

import pycro.backend.codegen as cgen
import pycro.backend.sourcelookup as sl

def _add_sp(codegen, offset):
    """ Remove a number of items from the top of the stack """
    assert offset >= 0
    if offset != 0:
        ADD = 0x83
        codegen.add_rm_instr(ADD, cgen.Mod.DIRECT_REGISTER, 0, cgen.Register.SP)
        codegen.add_instr(offset)

class Function:
    """ Class used to create a new function. The address of the
        function will be pushed onto the stack. """
    def __init__(self, codegen, name="<unknown>", filename="<unknown>"):
        self._codegen = codegen
        self._num_args = None
        self._references = []
        self._has_begun = False
        self._name = name
        self._filename = filename
        self._codegen.add_final_step(self._assert_was_begun)
        self._codegen.add_final_step(self._resolve)
        self._func_offset = None

    def _assert_was_begun(self):
        assert self._has_begun, "Function '%s' was never generated!" % self._name

    def add_reference(self):
        # Push a 'null' function onto the stack. When this function
        # is created, we will overwrite that value with the correct
        # offset.
        self._references.append(self._codegen.get_code_offset() + 1)
        self._codegen.push_data(0x0)

    def _resolve(self):
        # Write the function offset to the original call position push
        assert self._func_offset is not None
        func_offset = cgen.BASE_ADDRESS + self._func_offset
        for code_offset in self._references:
            self._codegen.set_instr(code_offset, func_offset & 0xff) # LSB
            self._codegen.set_instr(code_offset + 1, func_offset >> 8) # MSB

    def begin(self, num_arguments):
        """ Start the implementation of this function """
        assert not self._has_begun
        self._has_begun = True

        # We can add this even for compiled functions, as areas that fail
        # _should_ fail after some code has happend to compile.
        sl.add_source_reference(self._codegen, "<builtin %s>" % self._name, -1)

        logging.debug("Starting function... %s (%s)" % (self._name, self._filename))
        self._num_args = num_arguments

        self._func_offset = self._codegen.get_code_offset()
        self._last_return_offset = None

        # We want to adjust where we start looking at the stack so that it
        # includes the arguments that have been pushed (which also includes
        # the return address). 
        self._original_stack_size = self._codegen.get_stack_size()
        self._codegen.set_stack_size(self._original_stack_size + self._num_args * 2 + 2)

    def ret(self):
        self._codegen.pop(cgen.Register.AX)

        # Reset the stack to the position it was in at the start of the function. We don't
        # want to use the codegen 'pop_data' version, as that adjusts the state of the stack
        # (which we don't want as the func function can continue after the return statement)
        offset = self._codegen.get_stack_size() - (self._original_stack_size + self._num_args * 2 + 2)
        assert offset >= 0
        _add_sp(self._codegen, offset)

        RET = 0xC3
        self._codegen.add_instr(RET)
        self._last_return_offset = self._codegen.get_code_offset()

    def end(self):
        """ Finish the implementation of the function """
        # Make sure we added a 'return' as the last operation...
        assert self._codegen.get_code_offset() == self._last_return_offset
        self._codegen.set_stack_size(self._original_stack_size)
        logging.debug("Function compiled")


class Call:
    """ Call the function whose address is currently at the top of the stack. """
    def __init__(self, codegen):
        self._codegen = codegen

    def call(self, num_args):
        # Call the value currently on the stack
        MOV = 0x8B
        # BX = stack pointer
        self._codegen.add_rm_instr(MOV, cgen.Mod.DIRECT_REGISTER, cgen.Register.BX, cgen.Register.SP)


        CALL = 0xFF
        self._codegen.add_rm_instr(CALL, cgen.Mod.DISPLACE_ONE_BYTE, 2, cgen.R_M.BX, num_args * 2)

        # Pop the arguments, plus the function object
        num_bytes_to_pop = num_args * 2 + 2
        self._codegen.pop_data(num_bytes_to_pop)

        # Pop the arguments...
        self._codegen.push(cgen.Register.AX)
