from pycro.backend.codegen import ConditionCode, Register, Mod

class Integer:
    def __init__(self, codegen):
        self._codegen = codegen

    def _check_overflow(self):
        if not self._codegen.config.i386.overflow_checks:
            return

        import pycro.backend.operations.asserter as astr
        import pycro.backend.operations.string as string
        import pycro.backend.variable as var

        JCC = 0x80
        offset = self._codegen.get_code_offset()
        self._codegen.add_instr(0x0f)
        self._codegen.add_instr(JCC + ConditionCode.NO_OVERFLOW)
        self._codegen.add_instr(0)
        self._codegen.add_instr(0)

        string.add_string_reference(self._codegen, "Integer overflow!")
        message = var.StackVariable(self._codegen)
        astr.failure(self._codegen, message)
        self._codegen.pop_data()

        jump_size = self._codegen.get_code_offset() - offset - 4
        self._codegen.set_instr(offset + 2, jump_size & 0xFF)
        self._codegen.set_instr(offset + 3, (jump_size >> 8) & 0xFF)

    def add(self):
        self._codegen.add_binary_instr(0x01)
        self._check_overflow()

    def subtract(self, ignore_overflow=False):
        self._codegen.add_binary_instr(0x29)
        if not ignore_overflow:
            self._check_overflow()

    def multiply(self):
        """ Perform a signed multiply on the two items on the 
            top of the stack, pushing the result onto the stack """
        self._codegen.pop(Register.BX)
        self._codegen.pop(Register.AX)
        self._codegen.add_rm_instr(0xF7, Mod.DIRECT_REGISTER, 5, Register.BX)
        # Note that we are ignoring the high part of the result in DX for now...
        self._codegen.push(Register.AX)
        self._check_overflow()

    def _divide_by_zero_check(self):
        if not self._codegen.config.i386.overflow_checks:
            return

        import pycro.backend.control.label as lbl
        import pycro.backend.operations.asserter as astr
        import pycro.backend.operations.string as string
        import pycro.backend.variable as var
        divisor = var.StackVariable(self._codegen)

        divisor.push()
        self._codegen.push_data(0)
        self.equal()

        block = lbl.Label(self._codegen)
        block.jump_if_false()
        string.add_string_reference(self._codegen, "Divide by zero!")
        message = var.StackVariable(self._codegen)
        astr.failure(self._codegen, message)
        self._codegen.pop_data()
        block.resolve()

    def divide(self):
        """ Perform a signed divide on the two items on the 
            top of the stack, pushing the result onto the stack """
        self._divide_by_zero_check()

        self._codegen.pop(Register.BX)
        self._codegen.pop(Register.AX)
        # We need to sign extend AX into DX
        CWD = 0x99
        self._codegen.add_instr(CWD)

        # AX = DX:AX / BX
        self._codegen.add_rm_instr(0xF7, Mod.DIRECT_REGISTER, 7, Register.BX)
        self._codegen.push(Register.AX)

    def mod(self):
        """ Perform a signed divide on the two items on the 
            top of the stack, pushing the remainder onto the stack """
        self._codegen.pop(Register.BX)
        self._codegen.pop(Register.AX)
        self._codegen.push_data(0)
        self._codegen.pop(Register.DX)

        # DX = DX:AX % BX
        self._codegen.add_rm_instr(0xF7, Mod.DIRECT_REGISTER, 7, Register.BX)
        self._codegen.push(Register.DX)

    def shift_left(self):
        self._codegen.pop(Register.CX)
        self._codegen.pop(Register.AX)
        self._codegen.add_rm_instr(0xD3, Mod.DIRECT_REGISTER, 4, Register.AX)
        self._codegen.push(Register.AX)

    def shift_right(self):
        self._codegen.pop(Register.CX)
        self._codegen.pop(Register.AX)
        self._codegen.add_rm_instr(0xD3, Mod.DIRECT_REGISTER, 5, Register.AX)
        self._codegen.push(Register.AX)

    def _add_cmp_instr(self, condition_code):
        # Compare the two items currently on the stack by performing
        # a subtract operation. We aren't interested in the result
        # (only the condition code), so pop the result off the stack.
        self.subtract(ignore_overflow=True)
        self._codegen.pop(Register.AX)

        # Set ax to zero (not very efficient, but hey...)
        self._codegen.push_data(0)
        self._codegen.pop(Register.AX)

        # Check the result of the subtract operation by pushing the
        # relevant condition code onto the stack
        self._codegen.add_instr(0x0f)
        self._codegen.add_rm_instr(0x90 + condition_code, Mod.DIRECT_REGISTER, 0, Register.AX)
        self._codegen.push(Register.AX)

    def greater(self):
        self._add_cmp_instr(ConditionCode.GREATER)

    def greater_equal(self):
        self._add_cmp_instr(ConditionCode.GREATER_EQUAL)

    def lesser(self):
        self._add_cmp_instr(ConditionCode.LESSER)

    def unsigned_lesser(self):
        self._add_cmp_instr(ConditionCode.CARRY)

    def lesser_equal(self):
        self._add_cmp_instr(ConditionCode.LESSER_EQUAL)

    def equal(self):
        self._add_cmp_instr(ConditionCode.EQUAL)

    def not_equal(self):
        self._add_cmp_instr(ConditionCode.NOT_EQUAL)
        
    def binary_and(self):
        self._codegen.add_binary_instr(0x21)

