import pycro.backend.codegen as cgen 
import pycro.backend.control.label as lbl
import pycro.backend.function as func
from pycro.backend.operations.integer import Integer
import pycro.backend.operations.allocator as alloc
import pycro.backend.variable as var

from pycro.frontend.python.builtins import compiled
import pycro.frontend.python.compiler as cmplr

"""
Lists
=====

The list data type is a pointer to a structure containing a size
followed by a pointer to the list data.

eg:
    typedef struct ListContents {
        int size;
        int alloc_size;
        int data[];
    }*ListType;

"""

class _List:
    """ Class to implement a 'construct list' function """
    def __init__(self, codegen):
        self._codegen = codegen
        self.function = func.Function(codegen, "<list constructor>")
        self._have_created = False

    @compiled
    def _init_values(source, dest):
        """ Initialise a destination list from a set of source values.

            These source values are initialised from the stack (and appear
            as a list), but are in reverse order. """
        i = 0
        length = len(source)
        while i < length:
            dest[i] = source[length - i - 1]
            i = i + 1

    def _allocate(self, size):
        self._codegen.push_data(var._VARIABLE_SIZE * 3)
        alloc.allocate(self._codegen)
        array = var.StackVariable(self._codegen)

        # Set the size
        size.push()
        array.push()
        self._codegen.pop(cgen.Register.BX)
        self._codegen.pop(cgen.Register.AX)

        MOV_TO_ADDR = 0x89
        self._codegen.add_rm_instr(MOV_TO_ADDR, cgen.Mod.DISPLACE_ZERO_BYTES, 
                cgen.Register.AX, cgen.R_M.BX) 
        # Set the allocated size
        self._codegen.add_rm_instr(MOV_TO_ADDR, cgen.Mod.DISPLACE_ONE_BYTE, 
                cgen.Register.AX, cgen.R_M.BX, 2) 

        # Allocate the buffer
        size.push()
        self._codegen.push_data(var._VARIABLE_SIZE)
        Integer(self._codegen).multiply()
        alloc.allocate(self._codegen)

        # Set the buffer variable
        array.push()
        self._codegen.pop(cgen.Register.BX)
        self._codegen.pop(cgen.Register.AX)
        self._codegen.add_rm_instr(MOV_TO_ADDR, cgen.Mod.DISPLACE_ONE_BYTE, 
                cgen.Register.AX, cgen.R_M.BX, 4) 
        return array

    @compiled
    def _copy(source, dest):
        """ A function to be compiled that copies between two lists, and sets the final value """
        i = 0
        while i < len(source):
            dest[i] = source[i]
            i = i + 1

    def _reallocate(self, size, array):
        # Allocate twice as much as we need (to avoid re-allocations). This
        # is especially relevant at the moment, as we don't re-use 'unallocated'
        # memory. If we allocated only the required size each time, appending
        # X items would waste (X * 2 -1) + (x*2-2)... (2) bytes. eg: 200 appends
        # would waste 40kB of memory.
        size.push()
        self._codegen.push_data(2)
        Integer(self._codegen).multiply()
        num_allocated_items = var.StackVariable(self._codegen)

        # allocate the new buffer
        num_allocated_items.push()
        self._codegen.push_data(var._VARIABLE_SIZE)
        Integer(self._codegen).multiply()
        alloc.allocate(self._codegen)
        buffer = var.StackVariable(self._codegen)

        # Create a 'pretend' list to copy to the values to
        buffer.push()
        size.push()
        size.push()
        self._codegen.push(cgen.Register.SP)
        dest = var.StackVariable(self._codegen)

        # Call the function to copy the values
        self._copy.add_reference(self._codegen)
        array.push()
        dest.push()
        func.Call(self._codegen).call(2)
        self._codegen.pop_data()

        # Replace the old buffer with the new buffer
        num_allocated_items.push()
        buffer.push()
        array.push()
        self._codegen.pop(cgen.Register.BX)
        self._codegen.pop(cgen.Register.AX)
        MOV = 0x89
        self._codegen.add_rm_instr(MOV, cgen.Mod.DISPLACE_ONE_BYTE, cgen.Register.AX, cgen.R_M.BX, 4)

        # Set the allocated size
        self._codegen.pop(cgen.Register.AX)
        self._codegen.add_rm_instr(MOV, cgen.Mod.DISPLACE_ONE_BYTE, 
                cgen.Register.AX, cgen.R_M.BX, 2) 

        # Cleanup the stack
        self._codegen.pop_data(6 * var._VARIABLE_SIZE)

    def append(self, function, variables):
        array = variables[0]
        value = variables[1]

        array.push()
        length(self._codegen)
        current_size = var.StackVariable(self._codegen)

        # Get the new size
        current_size.push()
        self._codegen.push_data(1)
        Integer(self._codegen).add()
        size = var.StackVariable(self._codegen)

        # If we don't have enough size in the exising buffer, create a new one
        # First, push the allocated size (not the number of items, the amount
        # we have allocated)
        array.push()
        self._codegen.pop(cgen.Register.BX)
        MOV_TO_REG = 0x8B
        self._codegen.add_rm_instr(MOV_TO_REG, cgen.Mod.DISPLACE_ONE_BYTE, cgen.Register.AX, cgen.R_M.BX, 2)
        self._codegen.push(cgen.Register.AX)
        # Compare it to the amount we need
        size.push()
        Integer(self._codegen).lesser_equal()
        label = lbl.Label(self._codegen)
        label.jump_if_false()
        self._reallocate(size, array)
        label.resolve()

        # Set the new size
        array.push()
        size.push()
        self._codegen.pop(cgen.Register.AX)
        self._codegen.pop(cgen.Register.BX)
        MOV = 0x89
        self._codegen.add_rm_instr(MOV, cgen.Mod.DISPLACE_ZERO_BYTES, 
                cgen.Register.AX, cgen.R_M.BX) 

        # Append the new value
        value.push()
        array.push()
        current_size.push()
        set_item(self._codegen)

        # We need a return value
        self._codegen.push_data(0)
        function.ret()

    def _create_list(self):
        """ Implement a function to create a list from the items 
            currently on the stack. """
        self.function.begin(1)
        source = var.StackVariable(self._codegen, 1)

        # Get the length of the destination list
        source.push()
        length(self._codegen)
        size = var.StackVariable(self._codegen)

        dest = self._allocate(size)
        
        # Call the 'copy values' function.
        self._init_values.add_reference(self._codegen)

        source.push()
        dest.push()
        func.Call(self._codegen).call(2)

        # Return the allocated array
        dest.push()
        self.function.ret()
        self.function.end()

    def __call__(self):
        if not self._have_created:
            self._have_created = True
            self._create_list()

def create_list(codegen, length):
    """ Create a list of a fixed size from items currently on the stack.
    
        The top of the stack is the number of items. The items underneath
        it are the values to initialise in the list.  The pointer is pushed
        onto the stack. """
    # Push pointer to the values on the stack
    codegen.push(cgen.Register.SP)
    # Push the size of the 'list'
    codegen.push_data(length)
    # Push the amount allocated
    codegen.push_data(length)
    stack_pos = codegen.get_stack_size()

    construct_list = codegen.get_post_step(_List)
    construct_list.function.add_reference()
    codegen.push_data(codegen.get_stack_size() - stack_pos + 2) 
    codegen.push(cgen.Register.SP)
    Integer(codegen).add()
    func.Call(codegen).call(1)

    # Pop the allocated list, and remove the temporary list items on the stack,
    # then put the allocated list back on the stack.
    codegen.pop(cgen.Register.AX)
    codegen.pop_data((length + 3) * var._VARIABLE_SIZE)
    codegen.push(cgen.Register.AX)

@compiled
def _check_array_bounds(list,  i):
    assert i >= 0
    assert i < len(list)

def _check_bounds(codegen, x, i):
    if not codegen.config.i386.bounds_checks:
        return

    _check_array_bounds.add_reference(codegen)
    x.push()
    i.push()
    func.Call(codegen).call(2)
    codegen.pop_data()

def get_item(codegen):
    """ Push the byte at offset specified on top of the stack from
        the list at two from the top of the stack. """
    x = var.StackVariable(codegen, 1)
    i = var.StackVariable(codegen, 0)
    _check_bounds(codegen, x, i)

    MOV = 0x8B

    # For 'push x[i]'
    codegen.pop(cgen.Register.BX) # i
    codegen.pop(cgen.Register.SI) # x

    # Get the pointer to the buffer
    codegen.add_rm_instr(MOV, cgen.Mod.DISPLACE_ONE_BYTE, cgen.Register.SI, cgen.R_M.SI, 4)

    # As every entry is two bytes, we need to multiply
    # 'i' by two (i = i << 1)
    SHL = 0xD1
    codegen.add_rm_instr(SHL, cgen.Mod.DIRECT_REGISTER, 4, cgen.Register.BX)

    # AX = byte [BX + SI]
    codegen.add_rm_instr(MOV, cgen.Mod.DISPLACE_ZERO_BYTES, cgen.Register.AX, cgen.R_M.BX_SI)

    codegen.push(cgen.Register.AX)

def set_item(codegen):
    x = var.StackVariable(codegen, 1)
    i = var.StackVariable(codegen, 0)
    _check_bounds(codegen, x, i)

    # For 'x[i] = value
    codegen.pop(cgen.Register.BX) # i
    codegen.pop(cgen.Register.SI) # x
    codegen.pop(cgen.Register.AX) # value

    # Get the pointer to the buffer
    MOV = 0x8B
    codegen.add_rm_instr(MOV, cgen.Mod.DISPLACE_ONE_BYTE, cgen.Register.SI, cgen.R_M.SI, 4)

    # As every entry is two bytes, we need to multiply
    # 'i' by two (i = i << 1)
    SHL = 0xD1
    codegen.add_rm_instr(SHL, cgen.Mod.DIRECT_REGISTER, 4, cgen.Register.BX)

    # byte [BX + SI] = AX
    MOV = 0x89
    codegen.add_rm_instr(MOV, cgen.Mod.DISPLACE_ZERO_BYTES, cgen.Register.AX, cgen.R_M.BX_SI)

def append(codegen, function, variables):
    codegen.get_post_step(_List).append(function, variables)

def length(codegen):
    """ Push the length of the string at the top of the stack """
    codegen.pop(cgen.Register.BX)
    MOV = 0x8B
    codegen.add_rm_instr(MOV, cgen.Mod.DISPLACE_ZERO_BYTES, cgen.Register.AX, cgen.R_M.BX)
    codegen.push(cgen.Register.AX)

