import pycro.backend.codegen as cgen
import pycro.backend.function as func
import pycro.backend.variable as var
from pycro.backend.operations.integer import Integer
from pycro.backend.operations.softwareinterrupt import SoftwareInterrupt

class _String:
    """ Class to resolve references to a single constant text entry """
    def __init__(self, codegen, text):
        self._codegen = codegen
        self._references = []
        self._offset = None
        self._text = text

    def add_reference(self):
        self._references.append(self._codegen.get_code_offset() + 1)
        self._codegen.push_data(0x0)

    def resolve(self):
        if self._offset is None:
            self._offset = cgen.BASE_ADDRESS + self._codegen.get_code_offset()

            assert len(self._text) < (2 << 16)
            (msb, lsb) = self._get_msb_lsb(len(self._text))
            # Set the size
            self._codegen.add_instr(lsb)
            self._codegen.add_instr(msb)
            # Set the total allocated size
            self._codegen.add_instr(lsb)
            self._codegen.add_instr(msb)

            # Create a pointer to the next byte
            (msb, lsb) = self._get_msb_lsb(cgen.BASE_ADDRESS + self._codegen.get_code_offset() + 2)
            self._codegen.add_instr(lsb)
            self._codegen.add_instr(msb)
            for char in self._text:
                # Little endian
                self._codegen.add_instr(ord(char))
                self._codegen.add_instr(0)

        # Now resolve all references to the string
        (msb, lsb) = self._get_msb_lsb(self._offset)
        for reference in self._references:
            self._codegen.set_instr(reference, lsb)
            self._codegen.set_instr(reference + 1, msb)
        self._references = []

    def _get_msb_lsb(self, offset):
        msb = offset >> 8
        lsb = offset & 0xff
        return (msb, lsb)


class _StringTable:
    """ A class to track references to all constant text entries """
    def __init__(self, codegen):
        self._codegen = codegen
        self._entries = {}

    def add_reference(self, text):
        """ Add a reference to the constant string in the code """
        if text not in self._entries:
            self._entries[text] = _String(self._codegen, text)
        self._entries[text].add_reference()

    def __call__(self):
        """ Called after all references to the string have been defined """
        for string in self._entries.values():
            string.resolve()

def add_string_reference(codegen, text):
    codegen.get_post_step(_StringTable).add_reference(text)

def _print_char(codegen, function, variables):
    char = variables[0]
    char.push()
    SoftwareInterrupt(codegen).print_chr()
    codegen.push_data(0)
    function.ret()

def _print_string(print_char, string):
    for char in string:
        print_char(char)

def print_string(codegen):
    """ Print the string at the top of the buffer """
    # We have to put the import in here because of a
    # cyclic import problem...
    from pycro.frontend.python.builtins import builder, compiled

    string_var = var.StackVariable(codegen)
    compiled(_print_string).add_reference(codegen)
    builder(['char'])(_print_char).add_reference(codegen)
    string_var.push()
    func.Call(codegen).call(2)
    # Pop both references to the string from the stack
    codegen.pop_data(2 * var._VARIABLE_SIZE)
