#!env python
import unittest

import pycro.backend as back
import pycro.backend.codegen as cgen
import pycro.backend.operations.allocator as alloc
import pycro.backend.operations.integer as intgr
import pycro.backend.operations.softwareinterrupt as si
import pycro.backend.operations.string as string
import pycro.backend.test.comfile as cf
import pycro.backend.variable as var

class TestAllocator(unittest.TestCase):
    """ NOTE: This test case tests the _implementation_ of the 
        allocator, not the interface. """
    def _run(self, gen):
        return cf.run(gen.create_output())[0]

    def test_zero_allocation(self):
        """ Allocate a zero sized block, then another. Ensure the
            second block gets the same address as the first. """
        gen = cgen.CodeGen()
        gen.push_data(0)
        alloc.allocate(gen)
        gen.push_data(5)
        alloc.allocate(gen)
        intgr.Integer(gen).equal()
        si.SoftwareInterrupt(gen).exit()

        self.assertEqual(1, self._run(gen))

    def test_block_allocation(self):
        gen = cgen.CodeGen()
        gen.push_data(10)
        alloc.allocate(gen)
        a = var.StackVariable(gen)

        gen.push_data(0)
        alloc.allocate(gen)
        b = var.StackVariable(gen)

        a.push()
        gen.push_data(10)
        intgr.Integer(gen).add()
        b.push()
        intgr.Integer(gen).equal()
        si.SoftwareInterrupt(gen).exit()
        self.assertEqual(1, self._run(gen))

    def test_out_of_memory(self):
        """ Test what happens when we attempt to allocate too much memory """
        gen = cgen.CodeGen()
        gen.push_data(25000)
        alloc.allocate(gen)
        string.add_string_reference(gen, "1st should succeed\n")
        string.print_string(gen)

        gen.push_data(25000)
        alloc.allocate(gen)
        string.add_string_reference(gen, "2nd should succeed\n")
        string.print_string(gen)

        gen.push_data(30000)
        alloc.allocate(gen)
        string.add_string_reference(gen, "Doh! That one should have failed...")
        string.print_string(gen)

        gen.push_data(0)
        si.SoftwareInterrupt(gen).exit()
        result = cf.run(gen.create_output())
        self.assertEqual("1st should succeed\n2nd should succeed\n<builtin <allocate>> ( -1 ): AssertionError: Out of memory!\n", result[1])
        self.assertEqual(1, result[0])

    def test_no_allocator_overflow(self):
        """ Test that we can still allocate more then 32767 when the overflow tests are on """
        config = back.default_config()
        config.i386.overflow_checks = True

        gen = cgen.CodeGen(config)
        gen.push_data(25000)
        alloc.allocate(gen)
        string.add_string_reference(gen, "1st should succeed\n")
        string.print_string(gen)

        gen.push_data(25000)
        alloc.allocate(gen)
        string.add_string_reference(gen, "2nd should succeed\n")
        string.print_string(gen)
        gen.push_data(0)
        si.SoftwareInterrupt(gen).exit()
        result = cf.run(gen.create_output())
        self.assertEqual("1st should succeed\n2nd should succeed\n", result[1])
        self.assertEqual(0, result[0])

if __name__ == "__main__":
    unittest.main()
