#!env python
import unittest
from pycro.backend import default_config
from pycro.backend.operations.softwareinterrupt import SoftwareInterrupt
from pycro.backend.operations.integer import Integer
from pycro.backend.codegen import CodeGen
from pycro.backend.test.comfile import write_and_run, run

class TestInteger(unittest.TestCase):

    def _do_binary_operation(self, operation, a, b):
        
        generator = CodeGen()

        generator.push_data(a)
        generator.push_data(b)
        operation(Integer(generator))
        SoftwareInterrupt(generator).exit()
        return write_and_run(generator.create_output())

    def test_add(self):
        self.assertEqual(6, self._do_binary_operation(Integer.add, 1, 5))
        self.assertEqual(13, self._do_binary_operation(Integer.add, 10, 3))

    def test_subtract(self):
        self.assertEqual(2, self._do_binary_operation(Integer.subtract, 5, 3))
        self.assertEqual(8, self._do_binary_operation(Integer.subtract, 10, 2))

    def test_shift(self):
        self.assertEqual(40, self._do_binary_operation(Integer.shift_left, 5, 3))
        self.assertEqual(5, self._do_binary_operation(Integer.shift_left, 5, 0))

        self.assertEqual(0x0f, self._do_binary_operation(Integer.shift_right, 0xf0, 4))
        self.assertEqual(0x04, self._do_binary_operation(Integer.shift_right, 0x08, 1))

    def test_multiply(self):
        self.assertEqual(8, self._do_binary_operation(Integer.multiply, 2, 4))
        self.assertEqual(180, self._do_binary_operation(Integer.multiply, 60, 3))

    def test_divide(self):
        self.assertEqual(2, self._do_binary_operation(Integer.divide, 4, 2))
        self.assertEqual(2, self._do_binary_operation(Integer.divide, 5, 2))
        self.assertEqual(7, self._do_binary_operation(Integer.divide, 21, 3))

    def test_mod(self):
        self.assertEqual(0, self._do_binary_operation(Integer.mod, 8, 2))
        self.assertEqual(1, self._do_binary_operation(Integer.mod, 10, 3))
        self.assertEqual(2, self._do_binary_operation(Integer.mod, 11, 3))

    def test_binary_and(self):
        self.assertEqual(0x03, self._do_binary_operation(Integer.binary_and, 0xf3, 0x03))
        self.assertEqual(0xf0, self._do_binary_operation(Integer.binary_and, 0xf3, 0xf0))

    def _compare(self, a, b, operation):
        
        generator = CodeGen()

        generator.push_data(a)
        generator.push_data(b)
        operation(Integer(generator))
        SoftwareInterrupt(generator).exit()
        return write_and_run(generator.create_output())

    def test_greater(self):
        self.assertEqual(1, self._compare(5, 3, Integer.greater))
        self.assertEqual(0, self._compare(3, 5, Integer.greater))
        self.assertEqual(0, self._compare(5, 5, Integer.greater))

    def test_gerater_equal(self):
        self.assertEqual(1, self._compare(5, 3, Integer.greater_equal))
        self.assertEqual(0, self._compare(3, 5, Integer.greater_equal))
        self.assertEqual(1, self._compare(5, 5, Integer.greater_equal))

    def test_lesser(self):
        self.assertEqual(0, self._compare(5, 3, Integer.lesser))
        self.assertEqual(1, self._compare(3, 5, Integer.lesser))
        self.assertEqual(0, self._compare(5, 5, Integer.lesser))

        self.assertEqual(0, self._compare(5, -5, Integer.lesser))
        self.assertEqual(1, self._compare(-5, 5, Integer.lesser))

    def test_unsigned_lesser(self):
        self.assertEqual(1, self._compare(3, 5, Integer.unsigned_lesser))
        self.assertEqual(0, self._compare(-5, 5, Integer.unsigned_lesser))
        self.assertEqual(1, self._compare(5, -5, Integer.unsigned_lesser))

    def test_lesser_equal(self):
        self.assertEqual(0, self._compare(5, 3, Integer.lesser_equal))
        self.assertEqual(1, self._compare(3, 5, Integer.lesser_equal))
        self.assertEqual(1, self._compare(5, 5, Integer.lesser_equal))

    def test_equal(self):
        self.assertEqual(0, self._compare(5, 3, Integer.equal))
        self.assertEqual(0, self._compare(3, 5, Integer.equal))
        self.assertEqual(1, self._compare(5, 5, Integer.equal))

    def test_add_overflow(self):
        config = default_config()
        config.i386.overflow_checks = True

        gen = CodeGen(config)
        gen.push_data(20000)
        gen.push_data(20000)
        Integer(gen).add()
        SoftwareInterrupt(gen).exit()
        result = run(gen.create_output())
        self.assertEqual(1, result[0])
        self.assertEqual("<unknown> ( -1 ): AssertionError: Integer overflow!\n", result[1])

    def test_multiply_overflow(self):
        config = default_config()
        config.i386.overflow_checks = True

        gen = CodeGen(config)
        gen.push_data(20000)
        gen.push_data(20000)
        Integer(gen).multiply()
        SoftwareInterrupt(gen).exit()
        result = run(gen.create_output())
        self.assertEqual(1, result[0])
        self.assertEqual("<unknown> ( -1 ): AssertionError: Integer overflow!\n", result[1])

    def test_no_overflow(self):
        config = default_config()
        config.i386.overflow_checks = True

        gen = CodeGen(config)
        gen.push_data(10)
        gen.push_data(2)
        Integer(gen).multiply()
        SoftwareInterrupt(gen).exit()
        result = run(gen.create_output())
        self.assertEqual(20, result[0])
        self.assertEqual("", result[1])

    def test_divide_by_zero(self):
        config = default_config()
        config.i386.overflow_checks = True

        gen = CodeGen(config)
        gen.push_data(8)
        gen.push_data(0)
        Integer(gen).divide()
        SoftwareInterrupt(gen).exit()
        result = run(gen.create_output())
        self.assertEqual(1, result[0])
        self.assertEqual("<unknown> ( -1 ): AssertionError: Divide by zero!\n", result[1])

if __name__ == '__main__':
    unittest.main()

