#!env python
import unittest

import pycro.backend as back
from pycro.backend.codegen import CodeGen
import pycro.backend.function as func
import pycro.backend.operations.integer as integer
import pycro.backend.operations.list as lst
import pycro.backend.operations.softwareinterrupt as si
import pycro.backend.variable as var
from pycro.backend.test.comfile import run

class TestList(unittest.TestCase):
    def test_length(self):
        """ Test that we can correctly query the length of a constructed list """
        # assert sys.exit(len([3, 7])) == 2
        generator = CodeGen()
        generator.push_data(3)
        generator.push_data(7)
        lst.create_list(generator, 2)
        lst.length(generator)
        si.SoftwareInterrupt(generator).exit()
        self.assertEqual(2, run(generator.create_output())[0])

        # assert sys.exit(len([1, 13, 7])) == 3
        generator = CodeGen()
        generator.push_data(1)
        generator.push_data(13)
        generator.push_data(7)
        lst.create_list(generator, 3)
        lst.length(generator)
        si.SoftwareInterrupt(generator).exit()
        self.assertEqual(3, run(generator.create_output())[0])

    def test_length_empty_list(self):
        # assert sys.exit(len([])) == 0
        gen = CodeGen()
        lst.create_list(gen, 0)
        lst.length(gen)
        si.SoftwareInterrupt(gen).exit()
        self.assertEqual(0, run(gen.create_output())[0])

    def test_querying_first_item(self):
        generator = CodeGen()
        generator.push_data(3)
        generator.push_data(7)
        lst.create_list(generator, 2)
        
        # Query the first item
        generator.push_data(0)
        lst.get_item(generator)
        si.SoftwareInterrupt(generator).exit()

        self.assertEqual(3, run(generator.create_output())[0])

    def test_querying_second_item(self):
        generator = CodeGen()
        generator.push_data(3)
        generator.push_data(7)
        generator.push_data(5)
        lst.create_list(generator, 3)
        
        # Query the second item
        generator.push_data(1)
        lst.get_item(generator)
        si.SoftwareInterrupt(generator).exit()

        self.assertEqual(7, run(generator.create_output())[0])

    def test_set_item(self):
        gen = CodeGen()
        gen.push_data(3)
        gen.push_data(5)
        lst.create_list(gen, 2)
        a = var.StackVariable(gen)

        # Set the second item
        gen.push_data(7)
        a.push()
        gen.push_data(1)
        lst.set_item(gen)

        # Add the two items together
        a.push()
        gen.push_data(0)
        lst.get_item(gen)
        a.push()
        gen.push_data(1)
        lst.get_item(gen)
        integer.Integer(gen).add() # 3 + 7
        si.SoftwareInterrupt(gen).exit()
        self.assertEqual(10, run(gen.create_output())[0])

    def test_list_with_global(self):
        # There was a bug where lists (eg: allocation) would overwrite globals...
        gen = CodeGen()

        # Create a global variable
        a = var.GlobalVariable(gen)
        gen.push_data(5)
        a.pop()

        # Create a list, and ensure it doesn't trash the global variable
        gen.push_data(3)
        gen.push_data(7)
        lst.create_list(gen, 2)

        a.push()
        si.SoftwareInterrupt(gen).exit()
        self.assertEqual(5, run(gen.create_output())[0])

    def test_append(self):
        # [3, 7].append(4)
        gen = CodeGen()
        gen.push_data(3)
        gen.push_data(7)
        lst.create_list(gen, 2)
        array = var.StackVariable(gen)

        append = func.Function(gen, 2)
        append.add_reference()
        array.push()
        gen.push_data(4)
        func.Call(gen).call(2)
        gen.pop_data()

        # list[1] + list[2]
        array.push()
        gen.push_data(1)
        lst.get_item(gen)
        array.push()
        gen.push_data(2)
        lst.get_item(gen)
        integer.Integer(gen).add()
        si.SoftwareInterrupt(gen).exit()

        # Create the 'append' function...
        append.begin(2)
        variables = [var.StackVariable(gen, 2), var.StackVariable(gen, 1)]
        lst.append(gen, append, variables)
        append.end()

        self.assertEqual(11, run(gen.create_output())[0])

    def test_bounds_check_negative_array_access(self):
        config = back.default_config()
        config.i386.bounds_checks = True

        gen = CodeGen(config)

        # x = [3, 7]
        gen.push_data(3)
        gen.push_data(7)
        lst.create_list(gen, 2)
        x = var.StackVariable(gen)

        # x[-1]
        x.push()
        gen.push_data(-1)
        lst.get_item(gen)
        si.SoftwareInterrupt(gen).exit()

        result = run(gen.create_output())
        self.assertTrue("list.py" in result[1], result[1])
        self.assertTrue("AssertionError" in result[1], result[1])
        self.assertEqual(1, result[0])

    def test_bounds_too_high_array_access(self):
        config = back.default_config()
        config.i386.bounds_checks = True

        gen = CodeGen(config)

        # x = [3, 7]
        gen.push_data(3)
        gen.push_data(7)
        lst.create_list(gen, 2)
        x = var.StackVariable(gen)

        # x[8]
        x.push()
        gen.push_data(8)
        lst.get_item(gen)
        gen.push_data(0)
        si.SoftwareInterrupt(gen).exit()

        result = run(gen.create_output())
        self.assertTrue("list.py" in result[1], result[1])
        self.assertTrue("AssertionError" in result[1], result[1])
        self.assertEqual(1, result[0])

if __name__ == "__main__":
    unittest.main()

