#!env python
import unittest
import pycro.backend.codegen as cgen
import pycro.backend.operations.softwareinterrupt as si
import pycro.backend.function as func
import pycro.backend.test.comfile as cfile
import pycro.backend.variable as var
import pycro.backend.operations.integer as integer

class TestFunction(unittest.TestCase):
    def test_return(self):
        """ Test a simple return function """
        
        generator = cgen.CodeGen()
        function = func.Function(generator)
        function.add_reference()
        func.Call(generator).call(0)
        si.SoftwareInterrupt(generator).exit()

        function.begin(0)
        generator.push_data(0x3)
        function.ret()
        function.end()
        self.assertEqual(3, cfile.write_and_run(generator.create_output()))

    def test_stack_reset(self):
        """ Test that the stack is correctly reset when the 
            function has multiple items on the stack. """
        
        generator = cgen.CodeGen()
        function = func.Function(generator)
        function.add_reference()
        func.Call(generator).call(0)
        si.SoftwareInterrupt(generator).exit()

        function.begin(0)
        generator.push_data(0x3)
        generator.push_data(0x7)
        function.ret()
        function.end()
        self.assertEqual(7, cfile.write_and_run(generator.create_output()))

    def test_pre_function_variable(self):
        """ Test that we can correctly reference variables defined
            before the function call """
        
        generator = cgen.CodeGen()
        generator.push_data(0x1)
        vara = var.StackVariable(generator)
        function = func.Function(generator)
        function.add_reference()
        func.Call(generator).call(0)
        varb = var.StackVariable(generator)
        vara.push()
        si.SoftwareInterrupt(generator).exit()

        function.begin(0)
        generator.push_data(0x3)
        generator.push_data(0x7)
        function.ret()
        function.end()
        self.assertEqual(1, cfile.write_and_run(generator.create_output()))

    def test_function_from_function(self):
        """ Test that we can call a function from a function """
        
        generator = cgen.CodeGen()
        functiona = func.Function(generator)
        functiona.add_reference()
        func.Call(generator).call(0)
        si.SoftwareInterrupt(generator).exit()

        functiona.begin(0)
        functionb = func.Function(generator)
        functionb.add_reference()
        func.Call(generator).call(0)
        functiona.ret()
        functiona.end()

        functionb.begin(0)
        generator.push_data(0x4)
        functionb.ret()
        functionb.end()

        self.assertEqual(4, cfile.write_and_run(generator.create_output()))

    def test_multiple_returns(self):
        """ Test that we handle multiple return statements """
        
        generator = cgen.CodeGen()
        function = func.Function(generator)
        function.add_reference()
        func.Call(generator).call(0)
        si.SoftwareInterrupt(generator).exit()

        function.begin(0)
        generator.push_data(3)
        function.ret()
        generator.push_data(5)
        function.ret()
        function.end()

        self.assertEqual(3, cfile.write_and_run(generator.create_output()))

    def test_args(self):
        """ Test a simple function with an argument """
        
        generator = cgen.CodeGen()
        function = func.Function(generator)
        function.add_reference()
        generator.push_data(3)
        func.Call(generator).call(1)
        si.SoftwareInterrupt(generator).exit()

        function.begin(1)
        # Note that we have an offset of two (to skip the return address)
        variable = var.StackVariable(generator, 1)
        generator.push_data(5)
        variable.push()
        integer.Integer(generator).add()
        function.ret()
        function.end()

        self.assertEqual(8, cfile.write_and_run(generator.create_output()))

if __name__ == "__main__":
    unittest.main()

