from pycro.backend.codegen import Register, Mod, R_M

# The size of a variable on the stack. When we have proper classes, this will
# have to be detected from the type of variable on the stack... At the moment
# is is the size of an integer in 16bit i386 code.
_VARIABLE_SIZE = 2

class StackVariable(object):
    """ A variable keeps track of data on the stack. """

    def __init__(self, codegen, stack_index = 0):
        """ Create a new variable from the value on top of the stack. The 
            parameter 'stack_index' is the item in the stack we are
            referencing, where 0 is the top of the current stack. """
        self._codegen = codegen
        self._stack_offset = self._codegen.get_stack_size() - stack_index * _VARIABLE_SIZE

    def push(self):
        """ Push the variable to the top of the stack """
        offset = self.get_sp_offset()

        MOV = 0x8B

        # BX = stack pointer
        self._codegen.add_rm_instr(MOV, Mod.DIRECT_REGISTER, Register.BX, Register.SP)

        # AX = *(BX + offset)
        self._codegen.add_rm_instr(MOV, Mod.DISPLACE_ONE_BYTE, Register.AX, R_M.BX, offset)

        self._codegen.push(Register.AX)

    def pop(self):
        """ Pop the variable from the top of the stack, and store it in the variable """
        # Load variable into ax
        self._codegen.pop(Register.AX)
        offset = self.get_sp_offset()

        MOV = 0x89

        # BX = stack pointer
        self._codegen.add_rm_instr(MOV, Mod.DIRECT_REGISTER, Register.SP, Register.BX)

        # *(BX + offset) = AX
        self._codegen.add_rm_instr(MOV, Mod.DISPLACE_ONE_BYTE, Register.AX, R_M.BX, offset)

    def get_sp_offset(self):
        """ Get the address (relative to the current SP register) of this
            variable """
        return self._codegen.get_stack_size() - self._stack_offset

class GlobalVariable:
    """ A variable that is independant of the current stack """
    def __init__(self, codegen, should_auto_register=True):
        self._references = []
        self._codegen = codegen
        if should_auto_register:
            codegen.add_global_variable(self)

    def push(self):
        self._references.append(self._codegen.get_code_offset() + 1)

        MOV_MEM_TO_AX = 0xA1
        self._codegen.add_instr(MOV_MEM_TO_AX)
        self._codegen.add_instr(0)
        self._codegen.add_instr(0)

        self._codegen.push(Register.AX)

    def pop(self):
        self._codegen.pop(Register.AX)

        self._references.append(self._codegen.get_code_offset() + 1)
        MOV_AX_TO_MEM = 0xA3
        self._codegen.add_instr(MOV_AX_TO_MEM)
        self._codegen.add_instr(0)
        self._codegen.add_instr(0)

    def resolve(self, address):
        msb = address >> 8
        lsb = address & 0xff
        for reference in self._references:
            self._codegen.set_instr(reference, lsb)
            self._codegen.set_instr(reference + 1, msb)

