import pycro.backend.function as func
from pycro.backend.builder import builder
import pycro.frontend.python.compiler as cmplr

class _Compiler:
    """ Class to compile python code on demand """
    def __init__(self, codegen):
        self._codegen = codegen
        self._to_compile = []
        self._functions = {}

    def add_reference(self, function):
        """ Add a reference to a compiled function. """
        if function not in self._functions:
            self._functions[function] = func.Function(self._codegen, function.func_name)
            self._to_compile.append(function)
        self._functions[function].add_reference()

    def __call__(self):
        while len(self._to_compile) > 0:
            pyfunc = self._to_compile.pop()
            function = self._functions[pyfunc]

            compiler = cmplr.Compiler(self._codegen)
            function.begin(pyfunc.func_code.co_argcount)
            compiler.decode(function, pyfunc.func_code)
            function.end()

def compiled(func):
    """ Mark a function as to be compiled.

    Adds a 'add_reference' attribute to the function object which
    can be called to add a reference to the compiled function. It
    will not actually be compiled unless 'add_reference' is called.

    The add_reference function takes a single codegen parameter, and
    returns the number of arguments of the function.
    """
    def add_reference(codegen):
        compiler = codegen.get_post_step(_Compiler)
        compiler.add_reference(func)
    func.add_reference = add_reference
    return func

_builders = {}
def named_builder(args, name=None):
    """ Mark a function as being capable of generating the given named builtin function """
    def decorator(func):
        assert callable(func)
        func_name = name
        if func_name is None:
            func_name = func.func_name
        _builders[func_name] = builder(args)(func)
        return func
    return decorator

_builtins = {}
def has_builtin(name):
    return name in _builtins or name in _builders

def get_builtin(name):
    """
    Get a builtin function.
    
    Returns a function instance with an 'add_reference' method.
    """
    if name in _builtins:
        return _builtins[name]
    return _builders[name]

def builtin(func):
    """ A decorator to mark a function as 'builtin'.
    
        The function will be directly compiled into the output. """
    _builtins[func.__name__] = compiled(func)


_builtin_attrs = {}
_attr_builders = {}
def has_builtin_attr(name):
    return name in _builtin_attrs or name in _attr_builders

def get_builtin_attr(name, codegen):
    """
    Get a builtin attribute.

    Returns a function instance that has an add_reference method.
    """
    if name in _builtin_attrs:
        return _builtin_attrs[name]
    return _attr_builders[name]

def builtin_attr(func):
    """ A decorator to mark a function as being a builtin attribute """
    _builtin_attrs[func.__name__] = (compiled(func), func.func_code.co_argcount)

def builtin_attr_generator(args, name=None):
    """ Mark a function as being capable of generating the given named builtin function """
    assert isinstance(args, list)
    def decorator(func):
        assert callable(func)
        func_name = name
        if func_name is None:
            func_name = func.func_name
        _attr_builders[func_name] = (builder(args)(func), len(args))
        return func
    return decorator

