import pycro.backend.sourcelookup as sl

class Code:
    """ Class to handle getting the code bytes """
    def __init__(self, codegen, code, line_offsets, filename):
        """ Construct a Code object.

            @param code A list of tuplies indentifying (absolute offset, byte code character)
            @param line_offsets A list of (offset, lineno)
            @param filename Name of the file we are compiling """
        self._codegen = codegen
        self._code = code
        self._line_offsets = list(line_offsets)
        self._filename = filename

    def get_file_line(self, offset):
        """ Get the filename and line of the opcode at the given offset """
        line = 0
        for (next_offset, next_line) in self._line_offsets:
            if next_offset > offset:
                break
            line = next_line
        return (self._filename, line)

    def modify_code(self, index, value):
        self._code[index] = (self._code[index][0], value)

    def pop_opcode(self):
        """
        Pop an opcode from the front of the code block.
        """
        (filename, line) = self.get_file_line(self.get_absolute_pos())
        (offset, opcode) = self._code[0]
        opcode = ord(opcode)
        del self._code[0]

        sl.add_source_reference(self._codegen, filename, line)
        return opcode

    def peek_index(self, offset):
        index = (ord(self._code[offset + 1][1]) << 8) + ord(self._code[offset][1])
        return index

    def pop_index(self, offset=0):
        index = self.peek_index(offset)
        del self._code[offset:offset+2]
        return index

    def empty(self):
        return len(self._code) == 0

    def __len__(self):
        return len(self._code)

    def remove(self, start, end=None):
        """ Remove items in the range [start, end).
        
            If end is None, [start, start + 1) is removed.
            @return A code object containing the removed code """
        if end is None:
            end = start + 1
        code = self._code[start:end]
        del self._code[start:end]
        return Code(self._codegen, code, self._line_offsets, self._filename)

    def peek(self, index):
        """ Peek at the bytecode at a position relative to the current code """
        return ord(self._code[index][1])

    def get_relative_pos(self, absolute):
        """ Get the index of a given absolute position """
        for index, (offset, code) in enumerate(self._code):
            if offset == absolute:
                return index
        raise Exception("Failed to find absolute pos '%i'!" % absolute)

    def get_absolute_pos(self):
        """
        Get the absolute position of the opcode at the front of the block.
        """
        assert not self.empty()
        return self._code[0][0]

