import dis
import os.path

import pycro.backend.function as function
import pycro.backend.control.label as lbl
from pycro.backend.lib.sys import exit as sys_exit
import pycro.backend.operations.string as string
import pycro.backend.variable as var
from pycro.frontend.python import default_config
import pycro.frontend.python.code as cd
import pycro.frontend.python.opcode as opcode

class MissingGlobalError(Exception):
    """ Raised when a global variable is referenced (that hasn't been declared) """
    pass

class MissingModuleError(Exception):
    """ Raised when we attempt to import a module that doesn't exist """
    pass

class _Module:
    """ Class to hold information on a module """
    def __init__(self, name, codegen):
        self.init_func = function.Function(codegen, name, "<module '%s' initialisation>" %  name)
        self._name = name
        self.has_been_decoded = False
        self._variables = {}
        self._codegen = codegen

    def get_global_variable(self, name):
        if name not in self._variables:
            if self.has_been_decoded:
                raise MissingGlobalError(self._name, name)
            else:
                # This module hasn't been imported yet. Add a temporary
                # 'marker'
                self._variables[name] = var.GlobalVariable(self._codegen)
        return self._variables[name]

    def mark_as_decoded(self, func_code):
        assert not self.has_been_decoded
        self.has_been_decoded = True

        # If another module has referenced items in this module, make
        # sure those variables exist.
        for name in self._variables:
            if name not in func_code.co_names:
                raise MissingGlobalError(self._name, name)

        # For all the global names, create a variable.
        for name in func_code.co_names:
            if name not in self._variables:
                self._variables[name] = var.GlobalVariable(self._codegen)

class Compiler:
    """ Class to recursively compile a set of functions """
    def __init__(self, codegen, config=default_config()):
        self._codegen = codegen
        self._include_paths = config.python.include_path

        self._modules = {}

        # Members with 'function level' decoding context
        self._variables = None
        self._func_code = None
        self._func = None

        # Members with 'module level' decoding context
        self._module_name = None
        self._undecoded_functions = []

        self._labels = {}
        self._next_resolve_offset = 0
        
    def add_function_reference(self, func_code):
        """ Add a reference to a function object to be decoded. """
        func = function.Function(self._codegen, func_code.co_name, func_code.co_filename)
        func.add_reference()
        self._undecoded_functions.append((func, func_code))

    def get_label(self, offset):
        """
        Get a label to a given code offset.
        
        A label is a position in code that can be jumped to.
        """
        if offset not in self._labels:
            self._labels[offset] = lbl.Label(self._codegen)
        return self._labels[offset]

    def resolve_label(self, offset):
        """
        Mark a label as being decoded to the current position.

        All previous opcode offsets will also be resolved to
        the current position.
        """
        assert offset >= self._next_resolve_offset
        while self._next_resolve_offset <= offset:
            self.get_label(self._next_resolve_offset).resolve()
            self._next_resolve_offset += 1

    def add_return(self):
        self._func.ret()

    def get_constant(self, index):
        """ Get a reference to a constant in the current function """
        return self._func_code.co_consts[index]

    def get_variable(self, index):
        """ Get a variable in the function currently being decoded """
        return self._variables[index]
        
    def get_name(self, index):
        """ Get a name in the currently compiling code object """
        return self._func_code.co_names[index]

    def import_module(self, name):
        """ Mark a module to be imported.
        
            The 'import' method for that module will be called. """
        # We need to import any 'parent' modules first...
        names = name.split('.')
        if len(names) > 1:
            self.import_module('.'.join(names[:-1]))

        if name not in self._modules:
            self._modules[name] = _Module(name, self._codegen)
        self._modules[name].init_func.add_reference()
        function.Call(self._codegen).call(0)

    def get_global_variable(self, name, module):
        """ Get a referenced global object.
        
            @param name The name of the variable
            @param module The module to get the variable from. If None, 
                will be from the current module."""
        if module is None:
            module = self._module_name
        if module is None:
            raise MissingGlobalError("Builtin module is referencing an unknown variable '%s'!" % name)
        assert module in self._modules
        return self._modules[module].get_global_variable(name)

    def decode(self, function, func_code):
        """ Decode a specific block of code. Functions & strings referenced
            by this block of code are not compiled. """
        assert self._next_resolve_offset == 0
        assert len(self._labels) == 0

        self._func = function
        code = cd.Code(self._codegen, list(enumerate(func_code.co_code)), dis.findlinestarts(func_code), func_code.co_filename)

        self._variables = [None] * func_code.co_nlocals
        for index in range(func_code.co_argcount):
            self._variables[index] = var.StackVariable(self._codegen, func_code.co_argcount - index)

        for index in range(func_code.co_argcount, func_code.co_nlocals):
            # Initialise all variables to zero. We have to push some data
            # onto the stack (so the variables we create are looking at
            # the correct location in the stack)
            self._codegen.push_data(0)
            self._variables[index] = var.StackVariable(self._codegen)

        self._func_code = func_code
        opcode.Opcodes(self._codegen).decode(self, code)

        self._next_resolve_offset = 0
        self._labels.clear()

    def decode_module(self, func_code):
        """ Decode a code block, using it's return value as the
            exit code. """
        self._module_name = "__main__"

        # We 'call' the module initialisation code. We use the return value
        # as the exit code from the program.
        sys_exit.add_reference(self._codegen)
        self.import_module(self._module_name)
        function.Call(self._codegen).call(1)

        assert len(self._modules) == 1
        # For this 'main' module, we know the code, so we add it explicitly.
        self._undecoded_functions.append((self._modules[self._module_name].init_func, func_code))
        self._modules[self._module_name].mark_as_decoded(func_code)
        self._decode_undecoded_functions()

        # Decode any other modules that have been imported
        while 1:
            for name in self._modules:
                if not self._modules[name].has_been_decoded:
                    self._recursive_import(name.split('.'), self._include_paths)
                    break
            else:
                # No more modules to decoded
                break

    def _recursive_import(self, names, search_paths, name_prefix=""):
        """ Recursively import the item in the list names """
        assert len(names) > 0
        for include_path in search_paths:
            path = os.path.join(include_path, names[0])
            if os.path.isdir(path):
                path = os.path.join(path, "__init__.py")
            else:
                path = path + ".py"

            if os.path.exists(path):
                break
        else:
            raise MissingModuleError(name_prefix + names[0])

        self._perform_import(name_prefix + names[0], path)
        if len(names) > 1:
            prefix = "%s%s." % (name_prefix, names[0])
            path = os.path.join(include_path, names[0])
            self._recursive_import(names[1:], [path], prefix)

    def _perform_import(self, name, filename):
        """ Import a module from a file """
        if self._modules[name].has_been_decoded:
            return

        input_file = file(filename, "r")
        contents = input_file.read()
        input_file.close()
        
        code = compile(contents, filename, "exec")
        self._modules[name].mark_as_decoded(code)

        # Decode this module
        self._module_name = name
        self._undecoded_functions.append((self._modules[name].init_func, code))
        self._decode_undecoded_functions()

    def _decode_undecoded_functions(self):
        """ Decode all current functions """
        for (function, func_code) in self._undecoded_functions:
            function.begin(func_code.co_argcount)
            self.decode(function, func_code)
            function.end()
        self._undecoded_functions = []

