import dis
import logging

class OpcodeAllreadyRegisteredError(Exception):
    pass

log = logging.getLogger("pycro.frontend.python")

def _register_opcodes():
    import pycro.frontend.python.core
    import pycro.frontend.python.opcodes.assertion
    import pycro.frontend.python.opcodes.attributes
    import pycro.frontend.python.opcodes.compare
    import pycro.frontend.python.opcodes.constants
    import pycro.frontend.python.opcodes.forloop
    import pycro.frontend.python.opcodes.function
    import pycro.frontend.python.opcodes.globals
    import pycro.frontend.python.opcodes.jump
    import pycro.frontend.python.opcodes.list
    import pycro.frontend.python.opcodes.operations
    import pycro.frontend.python.opcodes.printitem
    import pycro.frontend.python.opcodes.variables
    import pycro.frontend.python.opcodes.whileloop

class _FilterLog:
    def filter(self, record):
        record.msg = " " * 4 + record.msg
        return 1

class _OpcodeDecoder:
    """ Class to hold information on creating a decoder for a certain type of opcode """
    def __init__(self, func):
        self._func = func

    def __call__(self, state, codegen, code, bcc):
        """ Call the decoder for the given opcode """
        self._func(state, codegen, code, bcc)

class Opcodes(object):
    """ Class to hold information about opcodes. Provides functionality
        for decoding code blocks using those opcodes. """
    # A map of opcodes to decoders
    _opcodes = {}

    def __init__(self, codegen):
        self._codegen = codegen
        _register_opcodes()

    @staticmethod
    def add_opcode(func, opcode_name):
        if opcode_name in Opcodes._opcodes:
            raise OpcodeAllreadyRegisteredError(opcode_name)

        Opcodes._opcodes[dis.opname.index(opcode_name)] = _OpcodeDecoder(func)

    def _decode(self, state, code):
        """ Decode a block of code """
        while not code.empty():
            offset = code.get_absolute_pos()

            # Resolve all jump references to this opcode
            state.resolve_label(offset)
            opcode = code.pop_opcode()

            if opcode not in self._opcodes:
                (filename, line) = code.get_file_line(offset)
                raise Exception("Opcode '%s' not handled (%s:%i)!" % (dis.opname[opcode], filename, line))

            log.debug("%s: (stack size before opcode = %i)", dis.opname[opcode], self._codegen.get_stack_size())
            try:
                self._opcodes[opcode](state, self._codegen, code, offset)
            except:
                (filename, line) = code.get_file_line(offset)
                logging.error("Exception while decoding %s[%i]" % (filename,  line))
                raise
            log.debug("\t (stack size after opcode = %i)", self._codegen.get_stack_size())

    def decode(self, state, code):
        """ Decode a block of code. Logging within this block will have whitespace
            leading whitespace added. """
        filter = _FilterLog()
        log.addFilter(filter)
        self._decode(state, code)
        log.removeFilter(filter)

