import dis

import pycro.frontend.python.builtins as builtins
from pycro.frontend.python.opcodes import opcode_with_index

class UnknownAttributeError (Exception):
    pass

@opcode_with_index('LOAD_ATTR')
def load_attribute(state, codegen, index, code, bcc):
    # We don't handle attributes, as this requires type analysis (eg:
    # we need to know the type of the item on the top of the stack, and
    # we need to know when attributes it has). What we can do is hard
    # code certain attributes that we need, and assume the type...
    #
    # One other problem is that the function call doesn't include the
    # instance in the number of arguments...
    name = state.get_name(index)

    if builtins.has_builtin_attr(name):
        (builtin, num_args) = builtins.get_builtin_attr(name, codegen)

        for i in range(num_args - 1):
            valid_opcodes = [dis.opname.index('LOAD_FAST'), dis.opname.index('LOAD_GLOBAL'), dis.opname.index('LOAD_CONST')]
            if code.peek(i * 3) not in valid_opcodes:
                (filename, line) = code.get_file_line(code.get_absolute_pos())
                raise NotImplementedError('%s(%i): Error: Method arguments can only be variable references (arg %i is an expression)!' % (filename, line, i))

        call_offset = num_args * 3 - 3
        if code.peek(call_offset) != dis.opname.index('CALL_FUNCTION'):
            (filename, line) = code.get_file_line(code.get_absolute_pos())
            raise NotImplementedError('%s(%i): Error: Cannot keep references to instance attributes (can only call them)!', (filename, line))

        # Modify the function call to have one more parameter (the instance parameter)
        code.modify_code(call_offset + 1, chr(code.peek_index(call_offset + 1) + 1))

        builtin.add_reference(codegen)

        # As we only get the function address from the get attribute, we need to
        # leave the instance on the stack to be called with. However, it must be
        # 'on top' of the function call...
        codegen.swap_top_of_stack()
    else:
        raise UnknownAttributeError(name)

