import dis
import types

import pycro.backend.operations.string as string
import pycro.frontend.python.builtins as builtin
from pycro.frontend.python.opcodes import opcode_with_index

class BuiltinFunctionError(Exception):
    pass

def _custom_import(state, codegen, code):
    # We are trying to do  'from X import Y'. The python opcode includes a
    # dummy LOAD_CONST tuple at the top level... note that what we are doing
    # is highly compiler dependant! If we handled tuples & other variable sized
    # data we wouldn't need to do this, as the push & pop of the tuple would 
    # 'just work'.
    assert code.peek(0) == dis.opname.index('IMPORT_NAME')
    code.pop_opcode()
    module_offset = code.pop_index()

    # Get all of the functions to import. Keep the 'store'
    # instructions to be processed normally.
    function_offsets = []
    index = 0
    while code.peek(index) == dis.opname.index('IMPORT_FROM'):
        code.remove(index)
        function_offsets.append(code.pop_index(index))

        assert code.peek(index) in [dis.opname.index('STORE_FAST'), dis.opname.index('STORE_NAME')]
        index = index + 3

    # Attempt to load builtin imports. We attempt to load builtins before
    # the local modules (as opposed to normal python) as we want to be able
    # to override our own custom modules (eg: the system window module).
    function_names = [state.get_name(offset) for offset in function_offsets]
    module_name = state.get_name(module_offset)

    # Push the 'imports' in reverse order, so that the 'stores' pick
    # up the correct function (as the first store to be run should get
    # the last function to be pushed).
    function_names.reverse()
    if not _import_builtin(module_name, function_names, codegen):
        _import_source_module(module_name, function_names, state)

    # Remove the reference to the popping of the unnecessary tuple.
    assert code.peek(index) == dis.opname.index('POP_TOP') 
    code.remove(index)

def _import_source_module(module_name, variables, compiler):
    """ Pushes the values of the named variables in the given module
        on the stack. """
    compiler.import_module(module_name)
    for name in variables:
        variable = compiler.get_global_variable(name, module_name)
        variable.push()

def _import_builtin(module_name, functions, codegen):
    """ Attempt to load a builtin module. Returns a boolean indicating
        whether we found the builtin module. """
    full_module_name = "pycro.backend.lib.%s" % module_name
    try:
        module = __import__(full_module_name)
    except ImportError:
        return False

    for comp in full_module_name.split('.')[1:]:
        module = getattr(module, comp)

    for function_name in functions:
        if not hasattr(module, function_name):
            raise BuiltinFunctionError("Backend module '%s' doesn't implement '%s'" % (full_module_name, function_name))
        function = getattr(module, function_name)
        if not hasattr(function, "add_reference"):
            raise BuiltinFunctionError("Backend function '%s' isn't marked with the builder tag!" % function)
        function.add_reference(codegen)
    return True

@opcode_with_index('IMPORT_NAME')
def import_error(state, codegen, index, code, bcc):
    raise NotImplementedError("Direct importing of modules not supported! Use the 'from X import Y' statement.")

@opcode_with_index('LOAD_CONST')
def load_const(state, codegen, index, code, bcc):
    value = state.get_constant(index)
    if value is None:
        value = 0

    if isinstance(value, int):
        codegen.push_data(value)
    elif isinstance(value, types.CodeType):
        state.add_function_reference(value)
    elif isinstance(value, str):
        string.add_string_reference(codegen, value)
    elif isinstance(value, tuple) and \
            code.peek(0) == dis.opname.index('IMPORT_NAME'):
        _custom_import(state, codegen, code)
    else:
        raise Exception("Unhandled constant type '%s' with value '%s'!" % (type(value), value))
