#!env python
import os.path
import shutil
import unittest

from pycro.frontend.python.compiler import MissingGlobalError
from pycro.frontend.python.test.run import run

class TestImport(unittest.TestCase):
    def _save(self, dir, module_name, contents):
        modules = module_name.split('.')
        for module in modules[0:-1]:
            dir = os.path.join(dir, module)
            if not os.path.exists(dir):
                os.mkdir(dir)
            file(os.path.join(dir, "__init__.py"), "w").close()
            
        temp = file(os.path.join(dir, "%s.py" % modules[-1]), "w")
        temp.write(contents)
        temp.close()

    def _run(self, main, *args):
        dir = os.path.join(os.path.split(__file__)[0], "temp")
        if os.path.exists(dir):
            shutil.rmtree(dir)
        os.mkdir(dir)

        for module, contents in args:
            self._save(dir, module, contents)

        return run(compile(main, "<test>", "exec"), [dir])
         
    def test_import_value_type(self):
        main = """from bob import cat
from sys import exit
exit(cat)"""

        self.assertEqual(3, self._run(main, ("bob", "cat = 3")))
        
    def test_import_function(self):
        main = """from bob import cat
from sys import exit
exit(cat())"""
        self.assertEqual(5, self._run(main, ("bob", "def cat(): return 5")))

    def test_sub_module_import(self):
        main = """from bob.rabbit import cat
from sys import exit
exit(cat())"""
        self.assertEqual(7, self._run(main, ("bob.rabbit", "def cat(): return 7")))

    def test_import_missing_name(self):
        main = """from bob import cat"""
        self.assertRaises(MissingGlobalError, self._run, main, ("bob", ""))

    def test_stack_after_import(self):
        """ Test that the stack is ok after an import """
        main = """i = 6
from bob import cat
from sys import exit
exit(i - cat)"""
        self.assertEqual(4, self._run(main, ("bob", "cat = 2")))

    def test_import_multiple_vars(self):
        main = """i = 6
from bob import a, b
from sys import exit
exit(b - a)"""
        self.assertEqual(2, self._run(main, ("bob", "a = 3\nb=5")))

    def test_submodule_from_module(self):
        main = """
from bugs.arron import apple
from sys import exit
exit(apple)"""
        arron = """
from bugs.bunny import bay
apple = bay"""
        bob = "bay = 11"
        self.assertEqual(11, self._run(main, ("bugs.arron", arron), ("bugs.bunny", bob)))

    def test_builtin_during_module(self):
        main = """
blah = "chicken"
from sys import exit
exit(len(blah)) """
        self.assertEqual(7, self._run(main))

# Repeated imports don't work! Every time a module is imported,
# it runs...
#    def test_repeated_import(self):
#        main = """from bob import cat
#i = cat() # 1
#i = i + cat() # 1 + 2
#from bob import cat
#i = i + cat() # 3 + 3
#i = i + cat() # 6 + 4
#from sys import exit
#exit(i)"""
#        module = """i = 0
#def cat():
#    global i
#    i = i + 1
#    return i"""
#        self.assertEqual(10, self._run(main, ("bob", module)))
    
if __name__ == "__main__":
    unittest.main()

