from system.drawing.window import set_pixel, refresh

_res_x = 320
_res_y = 150

MAX_Z = 10000

# Section = [x, z, z_slope, colour, next_index]
X = 0
Z = 1
Z_SLOPE = 2
COLOUR = 3
NEXT_SECTION = 4

# This is the precision of the slope a 'fixed point' value
SLOPE_MULTIPLE = 10

# Due to the fact that we are dealing with numbers with fixed
# precision, parallel lines (ie: same z_slope) can go infront
# and behind each other. As we cannot detect an intersection
# point for these lines, we simply ignore all intersections
# less then a certain distance away.
_IGNORE_DELTA = 1

# A list of available sections
_unused_sections = None
_scanlines = []

def _create_section(x, z, z_slope, colour):
    global _unused_sections
    if _unused_sections is None:
        result = [0, 0, 0, 0, None]
    else:
        result = _unused_sections
        _unused_sections = result[NEXT_SECTION]

    result[X] = x
    result[Z] = z
    result[Z_SLOPE] = z_slope
    result[COLOUR] = colour
    result[NEXT_SECTION] = None
    return result

def _add_free_section(section):
    global _unused_sections
    section[NEXT_SECTION] = _unused_sections
    _unused_sections = section

def _initialise_scanlines(scanlines):
    assert len(scanlines) == 0
    i = 0
    while i < _res_y:
        section = _create_section(0, 0, 0, 0)
        scanlines.append(section)
        i = i + 1

def _reset_scanline():
    """ Set the scanlines to 'black' """
    y = 0
    while y < len(_scanlines):
        # Release the old sections
        old = _scanlines[y]
        while old is not None:
            next = old[NEXT_SECTION] 
            _add_free_section(old)
            old = next

        # Create a black section the width of the screen
        _scanlines[y] = _create_section(0, MAX_Z, 0, 0)
        _scanlines[y][NEXT_SECTION] = _create_section(_res_x, None, None, None)

        y = y + 1

def _get_lesser_section(section, x):
    """
    Get the last section with an x value less then our own.
    """
    next_section = section[NEXT_SECTION]

    while next_section[X] < x:
        if next_section[NEXT_SECTION] is None:
            return section

        section = next_section
        next_section = section[NEXT_SECTION]
    return section

def _intercept(previous, x1, z1, z_slope, colour, intercept_z_slope, last, next, final_distance, x, z):
    """
    Create a section where the two lines cross
    """
    assert final_distance >= 0
    if final_distance <= _IGNORE_DELTA:
        # Don't bother creating an intersection; it is not only
        # fraught with errors, but it is a little meaningless when
        # they are so close. Just use the 'next' X
        dx = x
        dz = z
    else:
        # Find where the z values are equal...
        # ie: 
        #   z = previous[Z_SLOPE] * (x - previous[X]) + previous[Z]
        #   z = z_slope * (x - x1) + z1
        #   previous[Z_SLOPE] * (x - previous[X]) + previous[Z] = z_slope * (x - x1) + z1
        #   x * (previous[Z_SLOPE] - z_slope) = 
        #          previous[Z_SLOPE] * previous[X] - z_slope * x1 + z1
        dx = (previous[Z_SLOPE] * previous[X] - z_slope * x1 + (z1 - previous[Z]) * SLOPE_MULTIPLE) / (previous[Z_SLOPE] - z_slope)
        dz = (dx - x1) * z_slope / SLOPE_MULTIPLE + z1
        
    result = _create_section(dx, dz, intercept_z_slope, colour)

    # FIXME: bug 1
    #assert last[X] <= result[X]
    #assert result[X] <= next[X]
    last[NEXT_SECTION] = result
    result[NEXT_SECTION] = next
    return result

def add_span(y, x1, z1, x2, z2, colour):
    """ Add a row of colour to the screen. """
    global _scanlines
    if y < 0:
        return
    if y >= _res_y:
        return

    x1 = max(x1, _scanlines[y][X])
    x2 = min(x2, _res_x)
    if x1 >= x2:
        return

    z_slope = (z2 - z1) * SLOPE_MULTIPLE / (x2 - x1)

    previous = _get_lesser_section(_scanlines[y], x1)
    next = previous[NEXT_SECTION]
    assert previous is not None
    assert next is not None

    should_free_previous = False
    is_visible = False
    if next[X] == x1:
        # The next item starts where we do...
        if z1 <= next[Z]:
            # We can replace the next item entirely
            foreground = _create_section(x1, z1, z_slope, colour)
            previous[NEXT_SECTION] = foreground
            foreground[NEXT_SECTION] = next[NEXT_SECTION]

            should_free_previous = True
            is_visible = True
        previous = next
        next = previous[NEXT_SECTION]
    else:
        if previous[X] == x1:
            # We start on the left hand side of the screen
            assert previous[X] == 0
            assert previous is _scanlines[y]
            if z1 <= previous[Z]:
                # We are replacing the first section in the scanline
                should_free_previous = True
                foreground = _create_section(x1, z1, z_slope, colour)
                foreground[NEXT_SECTION] = next
                _scanlines[y] = foreground
                is_visible = True
        else:
            # We start in the middle of a section; find the depth of the background
            # where we are, and check to see if it is visible
            background_z = previous[Z_SLOPE] * (x1 - previous[X]) / SLOPE_MULTIPLE + previous[Z]
            if z1 <= background_z:
                is_visible = True

                foreground = _create_section(x1, z1, z_slope, colour)
                foreground[NEXT_SECTION] = next
                previous[NEXT_SECTION] = foreground

    # Walk along intermediate items, replacing where necessary
    while next[X] < x2:
        z = z_slope * (next[X] - x1) / SLOPE_MULTIPLE + z1

        # Test to see if we were visible at the end of the previous
        # section.
        should_free_next = False
        z_prev = previous[Z_SLOPE] * (next[X] - previous[X]) / SLOPE_MULTIPLE + previous[Z]
        if is_visible:
            if z > z_prev:
                # We are no longer visible... detect intersection
                _intercept(previous, x1, z1, z_slope, previous[COLOUR], previous[Z_SLOPE], foreground, next, z - z_prev, next[X], z_prev)
                # The previous part of the background now becomes the intersection point...
                if should_free_previous:
                    _add_free_section(previous)
                previous = foreground[NEXT_SECTION]
                should_free_previous = False
                is_visible = False
        else:
            if z <= z_prev:
                # We are now visible again! Find out where we became visible,
                # and add an intersection point.
                foreground = _intercept(previous, x1, z1, z_slope, colour, z_slope, previous, next, z_prev - z, next[X], z)
                is_visible = True

        # Check to see if we are visible at the start of the new segment
        if is_visible:
            if z <= next[Z]:
                # We are still visible; we can remove the
                # current point (as it is hidden).
                foreground[NEXT_SECTION] = next[NEXT_SECTION]
                should_free_next = True
            else:
                # We are no longer visble.
                is_visible = False
        else:
            if z < next[Z]:
                # We weren't visible, but now we are!
                foreground = _create_section(next[X], z, z_slope, colour)
                foreground[NEXT_SECTION] = next[NEXT_SECTION]
                previous[NEXT_SECTION] = foreground
                should_free_next = True
                is_visible = True

        if should_free_previous:
            _add_free_section(previous)
        should_free_previous = should_free_next

        previous = next
        next = next[NEXT_SECTION]

    # We now need to add in our 'final' point.
    assert next[X] >= x2
    assert previous[X] <= x2
    background_z = previous[Z_SLOPE] * (x2 - previous[X]) / SLOPE_MULTIPLE + previous[Z]
    if z2 <= background_z:
        # Our final point is visible
        if next[X] == x2:
            # No need to create a final point, as the next point
            # allready starts at the correct location.
            current = next
        else:
            # We need to add a point to switch back to the 'previous' colour.
            current = _create_section(x2, background_z, previous[Z_SLOPE], previous[COLOUR])
            current[NEXT_SECTION] = next

        if is_visible:
            foreground[NEXT_SECTION] = current
        else:
            # We weren't visible, but now we are. Find out where
            # we started being visible, and add that point before
            # our final point.
            _intercept(previous, x1, z1, z_slope, colour, z_slope, previous, current, background_z - z2, x2, z2)
    else:
        # Our final point isn't visible.
        if is_visible:
            # We went from visible to hidden; find out where, and
            # add a point there.
            _intercept(previous, x1, z1, z_slope, previous[COLOUR], previous[Z_SLOPE], foreground, next, z2 - background_z, x2, z2)

    if should_free_previous:
        _add_free_section(previous)

def swap():
    """ Draw the scanlined graphics.
    
        Calls system.drawing.window.refresh. """
    y = 0
    for line in _scanlines:
        section = line
        next = section[NEXT_SECTION]
        while next is not None:
            x = section[X]
            colour = section[COLOUR]
            max_x = next[X]

            while x < max_x:
                set_pixel(x, y, colour)
                x = x + 1

            section = next
            next = section[NEXT_SECTION]
        y = y + 1
    _reset_scanline()
    refresh()

_initialise_scanlines(_scanlines)
_reset_scanline()
