#!env python

import unittest

import common.graphics as graph
from common.graphics import _reset_scanline, _scanlines, add_span
from common.graphics import _res_x, _res_y, X, COLOUR, MAX_Z, NEXT_SECTION

class TestGraphics(unittest.TestCase):

    def _index(self, list, item):
        for i, entry in enumerate(list):
            if entry is item:
                return i
        else:
            raise IndexError("Failed to find item %s" % item)

    def _allocate(self, *args, **kwargs):
        result = self._create_section(*args, **kwargs)
        self._allocated.append(result)
        return result

    def _add_free_section(self, section):
        i = self._index(self._allocated, section)
        del self._allocated[i]
        self._free_section(section)

    def setUp(self):
        self._allocated = []

        self._create_section = graph._create_section
        graph._create_section = self._allocate
        _reset_scanline()

        self._free_section = graph._add_free_section
        graph._add_free_section = self._add_free_section

    def tearDown(self):
        graph._create_section = self._create_section

        _reset_scanline()
        graph._add_free_section = self._free_section

        # Look for the sections we have allocated (but not freed); this
        # represents an effective memory leak for the compiled version,
        # as it doesn't handle reference counting.
        undeleted = "\n".join(str(section) for section in self._allocated)
        if len(undeleted) > 0:
            self.fail("There were undeleted sections!%s" % undeleted)

    def test_initial_scanlines(self):
        """ Test that when we reset correctly """
        _reset_scanline()
        self.assertEqual(_res_y, len(_scanlines))
        for section in _scanlines:
            self.assertEqual(0, section[X])
            self.assertEqual(0, section[COLOUR])
            last = section[NEXT_SECTION]
            self.assertTrue(last is not None)
            self.assertEqual(_res_x, last[X])
            self.assertEqual(None, last[NEXT_SECTION])

    def expect_sections(self, sections, expected_sections):
        section = sections
        for index, expected in enumerate(expected_sections):
            self.assertTrue(section is not None)
            self.assertEqual(expected[0], section[X], \
                    "%s\nx value doesn't match at %i: %s != %s" % (sections, index, expected[0], section[X]))
            # Note that we don't care about the last colour
            if index != len(expected_sections) - 1:
                self.assertEqual(expected[1], section[COLOUR], \
                        "%s\ncolour value doesn't match at %i: %s != %s" % (sections, index, expected[1], section[COLOUR]))
            section =  section[NEXT_SECTION]

        self.assertTrue(section is None, "got unexpected sections... %s" % section)

    def test_single_span(self):
        add_span(9, 10, 0, 300, 0, 15)
        self.expect_sections(_scanlines[9], [[0, 0], [10, 15], [300, 0], [320, 0]])

    def test_fully_overlapping(self):
        add_span(9, 10, 0, 300, 0, 15)
        add_span(9, 5, 0, 310, 0, 8)
        self.expect_sections(_scanlines[9], [[0, 0], [5, 8], [310, 0], [320, 0]])

    def test_left_overlapping(self):
        add_span(9, 10, 0, 300, 0, 15)
        add_span(9, 5, 0, 15, 0, 7)
        self.expect_sections(_scanlines[9], [[0, 0], [5, 7], [15, 15], [300, 0], [320, 0]])

    def test_right_overlapping(self):
        add_span(9, 10, 0, 300, 0, 15)
        add_span(9, 195, 0, 310, 0, 7)
        self.expect_sections(_scanlines[9], [[0, 0], [10, 15], [195, 7], [310, 0], [320, 0]])

    def test_left_edge_overlapping(self):
        add_span(9, -10, 0, 300, 0, 15)
        self.expect_sections(_scanlines[9], [[0, 15], [300, 0], [320, 0]])

    def test_right_edge_overlapping(self):
        add_span(9, 10, 0, 400, 0, 15)
        self.expect_sections(_scanlines[9], [[0, 0], [10, 15], [320, 0]])

    def test_entry_in_middle(self):
        add_span(9, 10, 0, 300, 0, 13)
        add_span(9, 70, 0, 200, 0, 12)
        self.expect_sections(_scanlines[9], [[0, 0], [10, 13], [70, 12], [200, 13], [300, 0], [320, 0]])

    def test_non_overlapping(self):
        add_span(9, 10, 0, 100, 0, 13)
        add_span(9, 150, 0, 200, 0, 12)
        self.expect_sections(_scanlines[9], [[0, 0], [10, 13], [100, 0], [150, 12], [200, 0], [320, 0]])

    def _count_unused_sections(self):
        count = 0
        section = graph._unused_sections
        while section is not None:
            count += 1
            section = section[NEXT_SECTION]
        return count

    def test_freed_sections(self):
        """ Test that when we overwrite a section, that section is added to the free list """
        # This test is implementation dependant; it is just there to
        # ensure we aren't relying on reference counted objects (as
        # the compiler doesn't support them).
        graph._unused_sections = None
        add_span(9, 10, 0, 100, 0, 15)
        add_span(9, 110, 0, 200, 0, 15)
        self.assertEqual(0, self._count_unused_sections())
        add_span(9, 5, 0, 310, 0, 8)

        self.assertEqual(3, self._count_unused_sections())

    def test_left_and_right_overlap(self):
        """ Test a span that overlaps on both the left and right of the screen """
        add_span(9, -10, 0, 400, 0, 15)
        self.expect_sections(_scanlines[9], [[0, 15], [320, 0]])

    def test_second_section_behind_first(self):
        """
        Test when the second section is behind the first (but overlaps on the ends)

          z=0             150__________200
          z=1 100___________.............___________250
        """
        add_span(9, 150, 0, 200, 0, 15)
        add_span(9, 100, 1, 250, 1, 7)
        self.expect_sections(_scanlines[9], [[0,0], [100, 7], [150, 15], [200, 7], [250, 0], [320, 0]])

    def test_second_section_intercepts_first_on_left(self):
        """
        Test the intersection of two lines.

          z=0       \_
                      \_
                        \_
                          \_
          z=100   100.......\________200
                             ..
                               ..
                                 ..
          z=200                    ..
        """
        add_span(9, 100, 100, 200, 100, 15)
        add_span(9, 100, 0, 200, 200, 7)
        self.expect_sections(_scanlines[9], [[0,0], [100,7], [150, 15], [200, 0], [320, 0]])

    def test_second_section_intercepts_first_on_right(self):
        """
        Test the intersection of two lines.

          z=0                      _/
                                 _/
                               _/
                             _/
          z=100   100_______/........200
                         ..
                       ..
                     ..
          z=200    ..
        """
        add_span(9, 100, 100, 200, 100, 15)
        add_span(9, 100, 200, 200, 0, 7)
        self.expect_sections(_scanlines[9], [[0,0], [100,15], [150, 7], [200, 0], [320, 0]])

    def test_final_z_slope(self):
        """ 
        There was a bug where the z-slope of the section after a span was the
        z-slope of the added span (instead of the previous slope).

          z=0             150
          z=1                151
          z=2 100___________..___________________200
          z=3                          180..190

        When the bug is present, the z=3 will be displayed, as it will think
        the [151,200) section will have a slope of one (instead of zero).
        """
        add_span(7, 100, 2, 200, 2, 14)
        add_span(7, 150, 0, 151, 1, 15)
        add_span(7, 180, 3, 190, 3, 16)
        self.expect_sections(_scanlines[7], [[0, 0], [100, 14], [150, 15], [151, 14], [200, 0], [320, 0]])

    def test_precision(self):
        """
        Test the case where two parallel lines can change depth
        at different times (due to rounding), making one go
        behind the other (causing the intersection detection to
        go strange).

        We test by adding two sections, which have the same
        slope (100 across, 1 down), but make the second appear
        infront, then behind, the first. We add a third span
        to force front/back detection while the second span
        is behind the first.
        
          z=1  90__100___________
          z=2             |......|___________190__200

          z=4               130......150

        """

        add_span(7, 130, 3, 150, 3, 4)

        add_span(7, 100, 1, 200, 2, 1)
        add_span(7, 90, 1, 190, 2, 2)

        self.expect_sections(_scanlines[7], [[0, 0], [90, 2], [190, 1], [200, 0], [320, 0]])

    def _validate(self, y):
        """ Validate the given scanline """
        last = _scanlines[y]
        while last[NEXT_SECTION] is not None:
            self.assertTrue(last in self._allocated, _scanlines[y])
            next = last[NEXT_SECTION]
            self.assertTrue(next in self._allocated, _scanlines[y])

            # FIXME: bug 1
            #self.assertTrue(last[X] < next[X], _scanlines[y] )
            self.assertTrue(last[X] <= next[X], _scanlines[y] )
            last = next

    def test_hides_then_becomes_visible(self):
        """
        Test we handle a case where a visible added section is
        hidden (due to an intersection with the previous), then
        shown again at the start of the next section.
        """
        add_span(7, 100, 10, 150, 0, 1)
        add_span(7, 90, 5, 200, 5, 2)
        self.expect_sections(_scanlines[7], [[0, 0], [90, 2], [125, 1], [150, 2], [200, 0], [320, 0]])

    def test_becomes_visible_then_hides(self):
        """
        Test we handle a case where an added section becomes
        visible (due to an intersection with the previous),
        then is hidden again.
        """
        add_span(7, 100, 0, 150, 10, 1)
        add_span(7, 140, 0, 200, 0, 2)
        add_span(7, 90, 5, 200, 5, 3)
        self.expect_sections(_scanlines[7], [[0, 0], [90, 3], [100, 1], [125, 3], [140, 2], [200, 0], [320, 0]])

    def test_brute_force(self):
        """
        Run a brute-force scanline test.

        Just attempt to render lots of scan-lines, and verify the
        state of the scanline at the end of each section.
        """
        NUM_RUNS = 4
        SCANLINES_PER_RUN = 100

        # We want the test to be repeatable, so we seed the random
        # with a constant value.
        import random
        random.seed(0)
        for run in range(NUM_RUNS):
            _reset_scanline()
            for sl in range(SCANLINES_PER_RUN):
                x1 = random.randint(-_res_x, 2 * _res_x)
                z1 = random.randint(0, MAX_Z)
                width = random.randint(0, 100)
                x2 = x1 + width
                z2 = random.randint(0, MAX_Z)
                colour = random.randint(0, 256)
                add_span(7, x1, z1, x2, z2, colour)

                self._validate(7)

    def test_no_backwards_intersection_from_first_point(self):
        """
        Test that we don't attempt an intersection from the first
        point.

        We do this by adding a line (from 20 to 40) that appears
        on top of the first sloped line, but below the 'next';
        we want to test that no intersection is created between
        the sloped line and the line from 20-40.

          z=1     \_           8______20
                    \_
          z=3         \_    6__....15
                        \_  
                          \_
                            ..
                              ..
                                ..
          z=10                    ..
        """
        add_span(7, 0, 1, 10, 10, 1)
        add_span(7, 8, 1, 20, 1, 2)
        add_span(7, 6, 3, 15, 3, 3)
        self.expect_sections(_scanlines[7], [[0, 1], [6, 3], [8, 2], [20, 0], [320, 0]])


if __name__ == "__main__":
    unittest.main()
