#!env python
import sys
import time
import wx

__all__ = ["init", "setpixel", "refresh_screen"]
_window_size = (320, 200)

class _Window(wx.Window):
    def __init__(self, parent, ID):
        wx.Window.__init__(self, parent, ID, style=wx.NO_FULL_REPAINT_ON_RESIZE)
        self.Bind(wx.EVT_PAINT, self._onPaint)
        self.Bind(wx.EVT_SIZE, self._onSize)

        self._initBuffer()
        self._pixels = {}

    def _initBuffer(self):
        size = self.GetClientSize()
        self._buffer = wx.EmptyBitmap(max(1,size.width), max(1,size.height))

        dc = wx.BufferedDC(None, self._buffer)
        dc.SetBackground(wx.Brush(self.GetBackgroundColour()))
        dc.Clear()

    def _onSize(self, event):
        self._initBuffer()

    def _onPaint(self, event):
        dc = wx.BufferedPaintDC(self, self._buffer)

    def setPixel(self, x, y, r, g, b):
        key = (r,g,b)
        if key not in self._pixels:
            self._pixels[key] = []
        self._pixels[key].append((x,y))

    def refresh(self):
        dc = wx.BufferedDC(None, self._buffer)
        dc.BeginDrawing()
        for key in self._pixels:
            (r,g,b) = key
            pen = wx.Pen(wx.Colour(r, g, b), 1, wx.SOLID)
            dc.SetPen(pen)
            for x,y in self._pixels[key]:
                dc.DrawPoint(x, y)
        dc.EndDrawing()
        self._pixels = {}

        self.Refresh(False)


class _Application(wx.App):
    """ Application to initialise a window capable of repainting, and
        handle the event loop. """

    def __init__(self):
        wx.App.__init__(self, redirect=False)

        self._event_loop = wx.EventLoop()
        wx.EventLoop.SetActive(self._event_loop)

        self.handle_messages()

    def OnInit(self):
        self._frame = wx.Frame(None, -1, "pacman", size=(800, 600),
                        style=wx.DEFAULT_FRAME_STYLE | wx.NO_FULL_REPAINT_ON_RESIZE)
        self._frame.Bind(wx.EVT_CLOSE, self._onCloseWindow)

        self.window = _Window(self._frame, -1)
        self._frame.Show()
        self._frame.SetClientSize(_window_size)
        return True

    def _onCloseWindow(self, event):
        sys.exit(0)

    def handle_messages(self, timeout_seconds = None):
        start_time = time.clock()
        while 1:
            while self._event_loop.Pending():
                self._event_loop.Dispatch()

            # Check to see if we should wait for a bit longer...
            if timeout_seconds is None or start_time + timeout_seconds < time.clock():
                break
            time.sleep(0.03)

    def refresh(self):
        """ Redraw the main window, and handle any windows messages (as necessary). """
        self.window.refresh()
        self.handle_messages()

app = None

def init():
    """ Initialise the screen to 640x480 """
    global app
    assert app is None
    app = _Application()

_colours = [(255, 0, 0), (0, 255, 0), (0, 0, 255), (255, 255, 0), (0, 255, 255), (255, 0, 255)]
def set_pixel(x, y, colour):
    """ Set the colour of the pixel at a given position """
    global app
    assert app is not None
    if not 0 <= x < _window_size[0]:
        raise ValueError("Output of bounds x value '%i' (max %i)" % (x, _window_size[0]))
    if not 0 <= y < _window_size[1]:
        raise ValueError("Output of bounds y value '%i' (max %i)" % (y, _window_size[1]))
    if colour == 0:
        (r,g,b) = (0,0,0)
    else:
        (r, g, b) = _colours[colour % len(_colours)]
    app.window.setPixel(x, y, r, g, b)

def refresh():
    """ Display all the pixels that have been set since the last update """
    global app
    assert app is not None
    app.refresh()

def close():
    global app
    assert app is not None
    app.window.Close()

def _main():
    import datetime
    import time
    import math
    init()
    start_time = datetime.datetime.now()
    x = 0
    while datetime.datetime.now() - start_time < datetime.timedelta(seconds=10):
        y = 100 + 100 * math.sin(x / 100.0 * 2 * 3.1415)
        set_pixel(x, y, 5)
        x = x + 1
        refresh()
        time.sleep(0.01)

if __name__ == "__main__":
    _main()
